from __future__ import annotations

from typing import Protocol, Self, assert_never, overload

from sonolus.script.internal.meta_fn import perf_meta_fn
from sonolus.script.record import Record
from sonolus.script.values import zeros
from sonolus.script.vec import Vec2, pnpoly


class Quad(Record):
    """A quad defined by its four corners.

    Usage:
        ```python
        Quad(bl: Vec2, tl: Vec2, tr: Vec2, br: Vec2)
        ```
    """

    bl: Vec2
    """The bottom-left corner of the quad."""

    tl: Vec2
    """The top-left corner of the quad."""

    tr: Vec2
    """The top-right corner of the quad."""

    br: Vec2
    """The bottom-right corner of the quad."""

    @classmethod
    def zero(cls) -> Quad:
        """Return a quad with all corners set to (0, 0).

        Returns:
            A new quad with all corners at the origin.
        """
        return cls(
            bl=Vec2.zero(),
            tl=Vec2.zero(),
            tr=Vec2.zero(),
            br=Vec2.zero(),
        )

    @classmethod
    def from_quad(cls, value: QuadLike, /) -> Quad:
        """Create a quad from a quad-like value."""
        return cls(
            bl=value.bl,
            tl=value.tl,
            tr=value.tr,
            br=value.br,
        )

    @property
    def center(self) -> Vec2:
        """The center of the quad."""
        return (self.bl + self.tr + self.tl + self.br) / 4

    @property
    def mt(self) -> Vec2:
        """The midpoint of the top edge of the quad."""
        return (self.tl + self.tr) / 2

    @property
    def mr(self) -> Vec2:
        """The midpoint of the right edge of the quad."""
        return (self.tr + self.br) / 2

    @property
    def mb(self) -> Vec2:
        """The midpoint of the bottom edge of the quad."""
        return (self.bl + self.br) / 2

    @property
    def ml(self) -> Vec2:
        """The midpoint of the left edge of the quad."""
        return (self.bl + self.tl) / 2

    def translate(self, translation: Vec2, /) -> Quad:
        """Translate the quad by the given translation and return a new quad."""
        return Quad(
            bl=self.bl + translation,
            tl=self.tl + translation,
            tr=self.tr + translation,
            br=self.br + translation,
        )

    def scale(self, factor: Vec2, /) -> Quad:
        """Scale the quad by the given factor about the origin and return a new quad."""
        return Quad(
            bl=self.bl * factor,
            tl=self.tl * factor,
            tr=self.tr * factor,
            br=self.br * factor,
        )

    def scale_about(self, factor: Vec2, /, pivot: Vec2) -> Quad:
        """Scale the quad by the given factor about the given pivot and return a new quad."""
        return Quad(
            bl=(self.bl - pivot) * factor + pivot,
            tl=(self.tl - pivot) * factor + pivot,
            tr=(self.tr - pivot) * factor + pivot,
            br=(self.br - pivot) * factor + pivot,
        )

    def scale_centered(self, factor: Vec2, /) -> Quad:
        """Scale the quad by the given factor about its center and return a new quad."""
        return Quad(
            bl=self.bl * factor,
            tl=self.tl * factor,
            tr=self.tr * factor,
            br=self.br * factor,
        ).translate(self.center * (Vec2(1, 1) - factor))

    def rotate(self, angle: float, /) -> Quad:
        """Rotate the quad by the given angle about the origin and return a new quad.

        Args:
            angle: The angle of rotation in radians. Positive angles rotate counterclockwise.

        Returns:
            A new quad rotated by the given angle.
        """
        return Quad(
            bl=self.bl.rotate(angle),
            tl=self.tl.rotate(angle),
            tr=self.tr.rotate(angle),
            br=self.br.rotate(angle),
        )

    def rotate_about(
        self,
        angle: float,
        /,
        pivot: Vec2,
    ) -> Quad:
        """Rotate the quad by the given angle about the given pivot and return a new quad.

        Args:
            angle: The angle of rotation in radians. Positive angles rotate counterclockwise.
            pivot: The pivot point for rotation.

        Returns:
            A new quad rotated about the pivot by the given angle.
        """
        return Quad(
            bl=self.bl.rotate_about(angle, pivot),
            tl=self.tl.rotate_about(angle, pivot),
            tr=self.tr.rotate_about(angle, pivot),
            br=self.br.rotate_about(angle, pivot),
        )

    def rotate_centered(self, angle: float, /) -> Quad:
        """Rotate the quad by the given angle about its center and return a new quad.

        Args:
            angle: The angle of rotation in radians. Positive angles rotate counterclockwise.

        Returns:
            A new quad rotated about its center by the given angle.
        """
        return self.rotate_about(angle, self.center)

    def permute(self, count: int = 1, /) -> Quad:
        """Perform a cyclic permutation of the quad's vertices and return a new quad.

        On a square, this operation is equivalent to rotating the square counterclockwise 90 degrees `count` times.

        Negative values of `count` are allowed and will rotate the quad clockwise.

        Args:
            count: The number of vertices to shift. Defaults to 1.

        Returns:
            The permuted quad.
        """
        count = int(count % 4)
        result = zeros(Quad)
        match count:
            case 0:
                result.bl @= self.bl
                result.tl @= self.tl
                result.tr @= self.tr
                result.br @= self.br
            case 1:
                result.bl @= self.br
                result.tl @= self.bl
                result.tr @= self.tl
                result.br @= self.tr
            case 2:
                result.bl @= self.tr
                result.tl @= self.br
                result.tr @= self.bl
                result.br @= self.tl
            case 3:
                result.bl @= self.tl
                result.tl @= self.tr
                result.tr @= self.br
                result.br @= self.bl
        return result

    def contains_point(self, point: Vec2, /) -> bool:
        """Check if the quad contains the given point.

        It is not guaranteed whether points on the edges of the quad are considered inside or outside.

        Args:
            point: The point to check.

        Returns:
            True if the point is inside the quad, False otherwise.
        """
        return pnpoly((self.bl, self.tl, self.tr, self.br), point)


class Rect(Record):
    """A rectangle defined by its top, right, bottom, and left edges.

    Usage:
        ```python
        Rect(t: float, r: float, b: float, l: float)
        ```
    """

    t: float
    """The top edge of the rectangle."""

    r: float
    """The right edge of the rectangle."""

    b: float
    """The bottom edge of the rectangle."""

    l: float
    """The left edge of the rectangle."""

    @classmethod
    def from_center(cls, center: Vec2, dimensions: Vec2) -> Rect:
        """Create a rectangle from its center and dimensions."""
        return cls(
            t=center.y + dimensions.y / 2,
            r=center.x + dimensions.x / 2,
            b=center.y - dimensions.y / 2,
            l=center.x - dimensions.x / 2,
        )

    @overload
    @classmethod
    def from_margin(cls, trbl: float, /) -> Rect: ...

    @overload
    @classmethod
    def from_margin(cls, tb: float, lr: float, /) -> Rect: ...

    @overload
    @classmethod
    def from_margin(cls, t: float, lr: float, b: float, /) -> Rect: ...

    @overload
    @classmethod
    def from_margin(cls, t: float, r: float, b: float, l: float, /) -> Rect: ...

    @classmethod
    def from_margin(cls, a: float, b: float | None = None, c: float | None = None, d: float | None = None, /) -> Self:
        """Create a rectangle based on margins (edge distances) from the origin.

        Compared to the regular [`Rect`][sonolus.script.quad.Rect] constructor, this method negates the bottom and
        left values, and supports shorthands when fewer than four arguments are provided.

        The following signatures are supported:

        - `from_margin(trbl)`: All margins set to `trbl`.
        - `from_margin(tb, lr)`: Top and bottom margins set to `tb`, left and right margins set to `lr`.
        - `from_margin(t, lr, b)`: Top margin set to `t`, left and right margins set to `lr`, bottom margin set to `b`.
        - `from_margin(t, r, b, l)`: Top, right, bottom, and left margins set to `t`, `r`, `b`, and `l` respectively.

        Usage:
            ```python
            Rect.from_margin(1)  # Rect(t=1, r=1, b=-1, l=-1)
            Rect.from_margin(1, 2)  # Rect(t=1, r=2, b=-1, l=-2)
            Rect.from_margin(1, 2, 3)  # Rect(t=1, r=2, b=-3, l=-2)
            Rect.from_margin(1, 2, 3, 4)  # Rect(t=1, r=2, b=-3, l=-4)
            ```
        """
        args = (a, b, c, d)
        match args:
            case (a, None, None, None):
                return cls(t=a, r=a, b=-a, l=-a)
            case (a, b, None, None):
                return cls(t=a, r=b, b=-a, l=-b)
            case (a, b, c, None):
                return cls(t=a, r=b, b=-c, l=-b)
            case (a, b, c, d):
                return cls(t=a, r=b, b=-c, l=-d)
            case _:
                assert_never(args)

    @property
    def w(self) -> float:
        """The width of the rectangle."""
        return self.r - self.l

    @property
    def h(self) -> float:
        """The height of the rectangle."""
        return self.t - self.b

    @property
    def bl(self) -> Vec2:
        """The bottom-left corner of the rectangle."""
        return Vec2(self.l, self.b)

    @property
    def tl(self) -> Vec2:
        """The top-left corner of the rectangle."""
        return Vec2(self.l, self.t)

    @property
    def tr(self) -> Vec2:
        """The top-right corner of the rectangle."""
        return Vec2(self.r, self.t)

    @property
    def br(self) -> Vec2:
        """The bottom-right corner of the rectangle."""
        return Vec2(self.r, self.b)

    @property
    def center(self) -> Vec2:
        """The center of the rectangle."""
        return Vec2((self.l + self.r) / 2, (self.t + self.b) / 2)

    @property
    def mt(self) -> Vec2:
        """The middle-top point of the rectangle."""
        return Vec2((self.l + self.r) / 2, self.t)

    @property
    def mr(self) -> Vec2:
        """The middle-right point of the rectangle."""
        return Vec2(self.r, (self.t + self.b) / 2)

    @property
    def mb(self) -> Vec2:
        """The middle-bottom point of the rectangle."""
        return Vec2((self.l + self.r) / 2, self.b)

    @property
    def ml(self) -> Vec2:
        """The middle-left point of the rectangle."""
        return Vec2(self.l, (self.t + self.b) / 2)

    def as_quad(self) -> Quad:
        """Convert the rectangle to a [`Quad`][sonolus.script.quad.Quad]."""
        return Quad(
            bl=self.bl,
            tl=self.tl,
            tr=self.tr,
            br=self.br,
        )

    def translate(self, translation: Vec2, /) -> Rect:
        """Translate the rectangle by the given translation and return a new rectangle."""
        return Rect(
            t=self.t + translation.y,
            r=self.r + translation.x,
            b=self.b + translation.y,
            l=self.l + translation.x,
        )

    def scale(self, factor: Vec2, /) -> Rect:
        """Scale the rectangle by the given factor about the origin and return a new rectangle."""
        return Rect(
            t=self.t * factor.y,
            r=self.r * factor.x,
            b=self.b * factor.y,
            l=self.l * factor.x,
        )

    def scale_about(self, factor: Vec2, /, pivot: Vec2) -> Rect:
        """Scale the rectangle by the given factor about the given pivot and return a new rectangle."""
        return Rect(
            t=(self.t - pivot.y) * factor.y + pivot.y,
            r=(self.r - pivot.x) * factor.x + pivot.x,
            b=(self.b - pivot.y) * factor.y + pivot.y,
            l=(self.l - pivot.x) * factor.x + pivot.x,
        )

    def scale_centered(self, factor: Vec2, /) -> Rect:
        """Scale the rectangle by the given factor about its center and return a new rectangle."""
        return Rect(
            t=self.t * factor.y,
            r=self.r * factor.x,
            b=self.b * factor.y,
            l=self.l * factor.x,
        ).translate(self.center * (Vec2(1, 1) - factor))

    def expand(self, expansion: Vec2, /) -> Rect:
        """Expand the rectangle by the given amount and return a new rectangle."""
        return Rect(
            t=self.t + expansion.y,
            r=self.r + expansion.x,
            b=self.b - expansion.y,
            l=self.l - expansion.x,
        )

    def shrink(self, shrinkage: Vec2, /) -> Rect:
        """Shrink the rectangle by the given amount and return a new rectangle."""
        return Rect(
            t=self.t - shrinkage.y,
            r=self.r - shrinkage.x,
            b=self.b + shrinkage.y,
            l=self.l + shrinkage.x,
        )

    def contains_point(self, point: Vec2, /) -> bool:
        """Check if the rectangle contains the given point.

        Args:
            point: The point to check.

        Returns:
            True if the point is inside the rectangle, False otherwise.
        """
        return self.l <= point.x <= self.r and self.b <= point.y <= self.t


class _QuadLike(Protocol):
    """A protocol for types that can be used as quads."""

    @property
    def bl(self) -> Vec2:
        """The bottom-left corner of the quad."""

    @property
    def tl(self) -> Vec2:
        """The top-left corner of the quad."""

    @property
    def tr(self) -> Vec2:
        """The top-right corner of the quad."""

    @property
    def br(self) -> Vec2:
        """The bottom-right corner of the quad."""


# PyCharm doesn't recognize attributes as satisfying the protocol.
type QuadLike = _QuadLike | Quad
"""A type that can be used as a quad."""


@perf_meta_fn
def flatten_quad(quad: QuadLike) -> tuple[float, float, float, float, float, float, float, float]:
    bl = quad.bl
    tl = quad.tl
    tr = quad.tr
    br = quad.br
    return (
        bl.x,
        bl.y,
        tl.x,
        tl.y,
        tr.x,
        tr.y,
        br.x,
        br.y,
    )



from __future__ import annotations

from enum import Enum
from types import EllipsisType, FunctionType, MethodType, ModuleType, NoneType, NotImplementedType, UnionType
from typing import TYPE_CHECKING, Annotated, Final, Literal, TypeVar, Union, get_origin

if TYPE_CHECKING:
    from sonolus.script.internal.value import Value


def validate_value[T](value: T) -> Value | T:
    if isinstance(value, Value):
        return value
    if isinstance(value, int | float):
        return Num._accept_(value)
    if isinstance(value, Enum):
        return validate_value(value.value)
    if id(value) in BUILTIN_IMPLS:
        return validate_value(BUILTIN_IMPLS[id(value)])
    if isinstance(value, type):
        if value in {int, float, bool}:
            return constant.BasicConstantValue.of(Num)
        return constant.BasicConstantValue.of(value)

    if hasattr(value, "_init_") and callable(value._init_):
        try:
            value._init_()
        except Exception as e:
            raise RuntimeError(f"Error initializing value {value}: {e}") from e

    value_type = type(value)
    if value_type in {
        generic.PartialGeneric,
        TypeVar,
        FunctionType,
        MethodType,
        str,
        ModuleType,
        NoneType,
        NotImplementedType,
        EllipsisType,
        super,
    }:
        return constant.BasicConstantValue.of(value)
    if value_type is tuple:
        return tuple_impl.TupleImpl._accept_(value)
    if value_type is dict:
        return dict_impl.DictImpl._accept_(value)
    if value_type in {set, frozenset}:
        from sonolus.script.containers import FrozenNumSet

        return FrozenNumSet.of(*value)
    if get_origin(value) in {Literal, Annotated, UnionType, Final, tuple, type}:
        return constant.BasicConstantValue.of(value)
    if value in {Literal, Annotated, Union}:
        return constant.TypingSpecialFormConstant.of(value)
    if value_type is sonolus_globals._GlobalPlaceholder:
        return value.get()
    if getattr(value, "_is_comptime_value_", False):
        return constant.BasicConstantValue.of(value)
    raise TypeError(f"Unsupported value: {value!r}")


from sonolus.script import globals as sonolus_globals
from sonolus.script.internal import constant, dict_impl, generic, tuple_impl
from sonolus.script.internal.value import Value
from sonolus.script.num import Num

BUILTIN_IMPLS = {}



import pytest

from sonolus.script.array import Array
from sonolus.script.containers import Box
from sonolus.script.debug import debug_log
from sonolus.script.internal.error import CompilationError
from tests.script.conftest import run_and_validate, run_compiled


def test_simple_generator():
    def fn():
        def gen():
            yield 1
            yield 2
            yield 3

        for i in gen():
            debug_log(i)

    run_and_validate(fn)


def test_generator_interspersed():
    def fn():
        def gen():
            debug_log(1)
            yield 1
            debug_log(2)
            yield 2
            debug_log(3)
            yield 3

        for i in gen():
            debug_log(i)

    run_and_validate(fn)


def test_generator_laziness():
    def fn():
        def gen():
            debug_log(1)
            yield 1
            debug_log(2)
            yield 2
            debug_log(3)
            yield 3

        iterator = gen()
        debug_log(0)
        for i in iterator:
            debug_log(i)

    run_and_validate(fn)


def test_generator_over_array():
    def fn():
        arr = Array(1, 2, 3, 4, 5)

        def gen():
            for i in arr:
                yield i * 2

        for i in gen():
            debug_log(i)

    run_and_validate(fn)


def test_generator_over_tuple():
    def fn():
        tup = (1, 2, 3, 4, 5)

        def gen():
            for i in tup:
                yield i * 2

        for i in gen():
            debug_log(i)

    run_and_validate(fn)


def test_generator_over_nested_array():
    def fn():
        arr = Array(Array(1, 2), Array(3, 4))

        def gen():
            for sub_arr in arr:
                for i in sub_arr:
                    yield i * 2

        for i in gen():
            debug_log(i)

    run_and_validate(fn)


def test_generator_over_nested_tuple():
    def fn():
        tup = ((1, 2), (3, 4))

        def gen():
            for sub_tup in tup:
                for i in sub_tup:
                    yield i * 2

        for i in gen():
            debug_log(i)

    run_and_validate(fn)


def test_generator_over_nested_array_breaks():
    def fn():
        arr = Array(Array(1, 2), Array(3, 4))

        def gen():
            for sub_arr in arr:
                for i in sub_arr:
                    yield i * 2
                yield 123

        iterator = gen()
        for i in iterator:
            debug_log(i)
            break
        for i in iterator:
            debug_log(i * 2)

    run_and_validate(fn)


def test_generator_over_nested_tuple_breaks():
    def fn():
        tup = ((1, 2), (3, 4))

        def gen():
            for sub_tup in tup:
                for i in sub_tup:
                    yield i * 2
                yield 123

        iterator = gen()
        for i in iterator:
            debug_log(i)
            break
        for i in iterator:
            debug_log(i * 2)

    run_and_validate(fn)


def test_generator_with_skipping_loop():
    def fn():
        def gen():
            for i in range(10):
                if i % 2 == 0:  # Skip even numbers
                    continue
                yield i

        for i in gen():
            debug_log(i)

    run_and_validate(fn)


def test_generator_with_any_true():
    def fn():
        def gen():
            debug_log(1)
            yield 0
            debug_log(2)
            yield 1
            debug_log(3)
            yield 2

        return any(gen())

    assert run_and_validate(fn)


def test_generator_with_all_true():
    def fn():
        def gen():
            debug_log(1)
            yield 1
            debug_log(2)
            yield 2
            debug_log(3)
            yield 3

        return all(gen())

    assert run_and_validate(fn)


def test_generator_with_any_false():
    def fn():
        def gen():
            debug_log(1)
            yield 0
            debug_log(2)
            yield 0
            debug_log(3)
            yield 0

        return any(gen())

    assert not run_and_validate(fn)


def test_generator_with_all_false():
    def fn():
        def gen():
            debug_log(1)
            yield 0
            debug_log(2)
            yield 0
            debug_log(3)
            yield 0

        return all(gen())

    assert not run_and_validate(fn)


def test_generator_with_sum():
    def fn():
        def gen():
            yield 1
            yield 2
            yield 3

        return sum(gen())

    assert run_and_validate(fn) == 6


def test_generator_with_max():
    def fn():
        def gen():
            yield 1
            yield 2
            yield 3

        return max(gen())

    assert run_and_validate(fn) == 3


def test_generator_with_min():
    def fn():
        def gen():
            yield 1
            yield 2
            yield 3

        return min(gen())

    assert run_and_validate(fn) == 1


def test_generator_with_max_default():
    def fn():
        def gen():
            yield 1
            yield 2
            yield 3

        return max(gen(), default=0)

    assert run_and_validate(fn) == 3


def test_generator_with_min_default():
    def fn():
        def gen():
            yield 1
            yield 2
            yield 3

        return min(gen(), default=0)

    assert run_and_validate(fn) == 1


def test_empty_generator_with_max_default():
    def fn():
        def gen():
            return
            yield 1

        return max(gen(), default=0)

    assert run_and_validate(fn) == 0


def test_empty_generator_with_min_default():
    def fn():
        def gen():
            return
            yield 1

        return min(gen(), default=0)

    assert run_and_validate(fn) == 0


def test_generator_with_max_key():
    def fn():
        def gen():
            yield 1
            yield 2
            yield 3

        return max(gen(), key=lambda x: -x)

    assert run_and_validate(fn) == 1


def test_generator_with_min_key():
    def fn():
        def gen():
            yield 1
            yield 2
            yield 3

        return min(gen(), key=lambda x: -x)

    assert run_and_validate(fn) == 3


def test_parallel_generator():
    def fn():
        def inner_gen():
            yield 1
            yield 2

        def outer_gen():
            yield from inner_gen()
            yield 3

        for i in outer_gen():
            debug_log(i)

    run_and_validate(fn)


def test_nested_generator():
    def fn():
        def outer_gen():
            def inner_gen():
                yield 1
                yield 2

            yield from inner_gen()
            yield 3

        for i in outer_gen():
            debug_log(i)

    run_and_validate(fn)


def test_generator_changing_closure_loop():
    def fn():
        x = 0

        def gen():
            yield x
            yield x

        for i in gen():
            debug_log(i)
            x += 1

        return 0

    with pytest.raises(CompilationError, match=r"Binding 'x' has been modified.*"):
        run_compiled(fn)


def test_generator_changing_closure_sequential():
    def fn():
        x = 0

        def gen():
            yield x
            yield x

        iterator = gen()
        for i in iterator:
            debug_log(i)
            break

        x = 2
        for i in iterator:
            debug_log(i)

        return 0

    with pytest.raises(CompilationError, match=r"Binding 'x' has been modified.*"):
        run_compiled(fn)


def test_generator_with_return_in_middle():
    def fn():
        def gen():
            yield 1
            return
            yield 3

        for i in gen():
            debug_log(i)

    run_and_validate(fn)


def test_generator_with_return_in_loop():
    def fn():
        def gen():
            for i in range(5):
                if i == 3:
                    return
                yield i

        for i in gen():
            debug_log(i)

    run_and_validate(fn)


def test_infinite_generator():
    def fn():
        def gen():
            while True:
                yield 1

        for i, v in enumerate(gen()):
            debug_log(v)
            if i >= 5:
                break

    run_and_validate(fn)


def test_generator_with_single_parameter():
    def fn():
        def gen(x):
            yield x
            yield x * 2
            yield x * 3

        for i in gen(5):
            debug_log(i)

    run_and_validate(fn)


def test_generator_with_multiple_parameters():
    def fn():
        def gen(x, y):
            yield x
            yield y
            yield x + y

        for i in gen(10, 20):
            debug_log(i)

    run_and_validate(fn)


def test_generator_with_parameter_and_loop():
    def fn():
        def gen(multiplier):
            for i in range(3):
                yield i * multiplier

        for i in gen(4):
            debug_log(i)

    run_and_validate(fn)


def test_generator_with_parameter_over_array():
    def fn():
        arr = Array(1, 2, 3)

        def gen(factor):
            for i in arr:
                yield i * factor

        for i in gen(3):
            debug_log(i)

    run_and_validate(fn)


def test_generator_with_default_parameter():
    def fn():
        def gen(x=7):
            yield x
            yield x * 2

        for i in gen():
            debug_log(i)

    run_and_validate(fn)


def test_generator_with_mixed_parameters():
    def fn():
        def gen(x, y=10):
            yield x
            yield y
            yield x * y

        for i in gen(5):
            debug_log(i)

    run_and_validate(fn)


def test_generator_with_parameter_in_nested_call():
    def fn():
        def inner_gen(value):
            yield value
            yield value + 1

        def outer_gen(base):
            yield from inner_gen(base)
            yield from inner_gen(base * 2)

        for i in outer_gen(3):
            debug_log(i)

    run_and_validate(fn)


def test_generator_yielding_record():
    def fn():
        def gen():
            for i in range(10):
                yield Box(i)

        for record in gen():
            debug_log(record.value)

    run_and_validate(fn)


def test_generator_yielding_record_mutation():
    def fn():
        box = Box(1)

        def gen():
            yield box
            yield box
            yield box

        for record in gen():
            debug_log(record.value)

        for record in gen():
            debug_log(record.value)
            box.value = 2

    run_and_validate(fn)


def test_generator_yielding_array_record_element_with_mutation():
    def fn():
        arr = Array(Box(1), Box(2), Box(3))

        def gen():
            yield from arr

        for record in gen():
            debug_log(record.value)
            record.value = 10

        for record in gen():
            arr[0].value = 20
            debug_log(record.value)

        for element in gen():
            debug_log(element.value)

    run_and_validate(fn)


def test_nested_iteration_of_same_generator():
    def fn():
        arr = Array(Box(0), Box(1), Box(2))

        def gen():
            for i in range(3):
                yield arr[i]

        for record in gen():
            for record_2 in gen():
                record.value = 1
                debug_log(record.value + record_2.value)

    run_and_validate(fn)


def test_generator_with_next():
    def fn():
        def gen():
            yield 1
            yield 2
            yield 3

        iterator = gen()
        debug_log(next(iterator))
        debug_log(next(iterator))
        debug_log(next(iterator))

    run_and_validate(fn)


def test_generator_with_iter():
    def fn():
        def gen():
            yield 1
            yield 2
            yield 3

        iterator = iter(gen())
        for i in iterator:
            debug_log(i)

    run_and_validate(fn)


def test_comptime_empty_generator():
    def fn():
        def gen():
            return
            yield lambda: 1

        for i in gen():
            return i()
        return 2

    assert run_and_validate(fn) == 2



from collections.abc import Callable, Sequence
from concurrent.futures import Executor, Future
from threading import Event, Lock

from sonolus.backend.finalize import cfg_to_engine_node
from sonolus.backend.ir import IRConst, IRInstr
from sonolus.backend.mode import Mode
from sonolus.backend.node import EngineNode
from sonolus.backend.ops import Op
from sonolus.backend.optimize.flow import BasicBlock, hash_cfg
from sonolus.backend.optimize.optimize import STANDARD_PASSES
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig, run_passes
from sonolus.backend.visitor import compile_and_call_at_definition
from sonolus.build.node import OutputNodeGenerator
from sonolus.script.archetype import _BaseArchetype
from sonolus.script.internal.callbacks import CallbackInfo
from sonolus.script.internal.context import (
    CallbackContextState,
    Context,
    ModeContextState,
    ProjectContextState,
    context_to_cfg,
    ctx,
    using_ctx,
)
from sonolus.script.internal.error import CompilationError
from sonolus.script.num import _is_num


class CompileCache:
    def __init__(self):
        self._cache = {}
        self._lock = Lock()
        self._event = Event()
        self._accessed_hashes = set()

    def get(self, entry_hash: int) -> EngineNode | None:
        with self._lock:
            self._accessed_hashes.add(entry_hash)
            if entry_hash not in self._cache:
                self._cache[entry_hash] = None  # Mark as being compiled
                return None
        while True:
            with self._lock:
                node = self._cache[entry_hash]
                if node is not None:
                    return node
            self._event.wait()

    def set(self, entry_hash: int, node: EngineNode) -> None:
        with self._lock:
            self._accessed_hashes.add(entry_hash)
            self._cache[entry_hash] = node
            self._event.set()
            self._event.clear()

    def reset_accessed(self) -> None:
        with self._lock:
            self._accessed_hashes.clear()

    def prune_unaccessed(self) -> None:
        with self._lock:
            unaccessed_hashes = set(self._cache.keys()) - self._accessed_hashes
            for h in unaccessed_hashes:
                del self._cache[h]


def compile_mode(
    mode: Mode,
    project_state: ProjectContextState,
    archetypes: list[type[_BaseArchetype]] | None,
    global_callbacks: list[tuple[CallbackInfo, Callable]] | None,
    passes: Sequence[CompilerPass] | None = None,
    thread_pool: Executor | None = None,
    validate_only: bool = False,
    cache: CompileCache | None = None,
) -> dict:
    if passes is None:
        passes = STANDARD_PASSES

    mode_state = ModeContextState(
        mode,
        archetypes,
    )
    nodes = OutputNodeGenerator()
    results = {}

    def process_callback(
        cb_info: CallbackInfo,
        cb: Callable,
        arch: type[_BaseArchetype] | None = None,
    ) -> tuple[str, int] | tuple[str, dict]:
        """Compile a single callback to a node.

        Returns either:
        - (cb_info.name, node_index) for global callbacks, or
        - (cb_info.name, {"index": node_index, "order": cb_order}) for archetype callbacks.
        """
        cfg = callback_to_cfg(project_state, mode_state, cb, cb_info.name, arch)
        if validate_only:
            if arch is not None:
                cb_order = getattr(cb, "_callback_order_", 0)
                return cb_info.name, {"index": 0, "order": cb_order}
            else:
                return cb_info.name, 0

        if cache is not None:
            cfg_hash = hash_cfg(cfg)
            node = cache.get(cfg_hash)
            if node is None:
                optimized_cfg = run_passes(cfg, passes, OptimizerConfig(mode=mode, callback=cb_info.name))
                node = cfg_to_engine_node(optimized_cfg)
                cache.set(cfg_hash, node)
            node_index = nodes.add(node)
        else:
            optimized_cfg = run_passes(cfg, passes, OptimizerConfig(mode=mode, callback=cb_info.name))
            node = cfg_to_engine_node(optimized_cfg)
            node_index = nodes.add(node)

        if arch is not None:
            cb_order = getattr(cb, "_callback_order_", 0)
            return cb_info.name, {"index": node_index, "order": cb_order}
        else:
            return cb_info.name, node_index

    all_futures = {}
    base_archetype_entries = {}

    if archetypes is not None:
        base_archetypes = []
        seen_base_archetypes = set()
        for a in archetypes:
            base = getattr(a, "_derived_base_", a)
            if base not in seen_base_archetypes:
                seen_base_archetypes.add(base)
                base_archetypes.append(base)

        for archetype in base_archetypes:
            archetype._init_fields()

            imports = []
            for name, import_info in archetype._imported_keys_.items():
                import_entry = {"name": name, "index": import_info.index}
                if import_info.default is not None:
                    import_entry["def"] = import_info.default
                imports.append(import_entry)

            archetype_data = {
                "name": archetype.name,
                "hasInput": archetype.is_scored,
                "imports": imports,
            }
            if mode == Mode.PLAY:
                archetype_data["exports"] = [*archetype._exported_keys_]

            callback_items = [
                (cb_name, cb_info, archetype._callbacks_[cb_name])
                for cb_name, cb_info in archetype._supported_callbacks_.items()
                if cb_name in archetype._callbacks_
                and archetype._callbacks_[cb_name] not in archetype._default_callbacks_
            ]

            if thread_pool is not None:
                for cb_name, cb_info, cb in callback_items:
                    cb_order = getattr(cb, "_callback_order_", 0)
                    if not cb_info.supports_order and cb_order != 0:
                        raise ValueError(f"Callback '{cb_name}' does not support a non-zero order")
                    f: Future = thread_pool.submit(process_callback, cb_info, cb, archetype)
                    all_futures[f] = ("archetype", archetype_data, cb_name)
            else:
                for cb_name, cb_info, cb in callback_items:
                    cb_order = getattr(cb, "_callback_order_", 0)
                    if not cb_info.supports_order and cb_order != 0:
                        raise ValueError(f"Callback '{cb_name}' does not support a non-zero order")
                    cb_name, result_data = process_callback(cb_info, cb, archetype)
                    archetype_data[cb_name] = result_data

            base_archetype_entries[archetype] = archetype_data

    if global_callbacks is not None and thread_pool is not None:
        for cb_info, cb in global_callbacks:
            f: Future = thread_pool.submit(process_callback, cb_info, cb, None)
            all_futures[f] = ("global", None, cb_info.name)

    if thread_pool is not None:
        for f, (callback_type, archetype_data, cb_name) in all_futures.items():
            cb_name, result_data = f.result()
            if callback_type == "archetype":
                archetype_data[cb_name] = result_data
            else:  # global callback
                results[cb_name] = result_data
    elif global_callbacks is not None:
        for cb_info, cb in global_callbacks:
            cb_name, node_index = process_callback(cb_info, cb, None)
            results[cb_name] = node_index

    if archetypes is not None:
        results["archetypes"] = [
            {**base_archetype_entries[getattr(a, "_derived_base_", a)], "name": a.name, "hasInput": a.is_scored}
            for a in archetypes
        ]

    results["nodes"] = nodes.get()
    return results


def callback_to_cfg(
    project_state: ProjectContextState,
    mode_state: ModeContextState,
    callback: Callable,
    name: str,
    archetype: type[_BaseArchetype] | None = None,
) -> BasicBlock:
    try:
        # Default to no_eval=True for performance unless there's an error.
        return _callback_to_cfg(project_state, mode_state, callback, name, archetype, no_eval=True)
    except CompilationError:
        return _callback_to_cfg(project_state, mode_state, callback, name, archetype, no_eval=False)


def _callback_to_cfg(
    project_state: ProjectContextState,
    mode_state: ModeContextState,
    callback: Callable,
    name: str,
    archetype: type[_BaseArchetype] | None,
    no_eval: bool,
) -> BasicBlock:
    callback_state = CallbackContextState(name, no_eval=no_eval)
    context = Context(project_state, mode_state, callback_state)
    with using_ctx(context):
        if archetype is not None:
            result = compile_and_call_at_definition(callback, archetype._for_compilation())
        else:
            result = compile_and_call_at_definition(callback)
        if _is_num(result):
            ctx().add_statements(IRInstr(Op.Break, [IRConst(1), result.ir()]))
    return context_to_cfg(context)



from __future__ import annotations

from collections.abc import Iterator

from sonolus.backend.blocks import Block

type Place = BlockPlace | SSAPlace
type BlockValue = Block | int | TempBlock | Place
type IndexValue = int | Place


class TempBlock:
    __slots__ = ("__hash", "name", "size")

    def __init__(self, name: str, size: int = 1):
        self.name = name
        self.size = size
        self.__hash = hash(name)  # Precompute hash based on name alone

    def __repr__(self):
        return f"TempBlock(name={self.name!r}, size={self.size!r})"

    def __str__(self):
        return f"{self.name}"

    def __getitem__(self, item) -> BlockPlace:
        return BlockPlace(self, item)

    def __iter__(self) -> Iterator[BlockPlace]:
        for i in range(self.size):
            yield self[i]

    def __eq__(self, other):
        return isinstance(other, TempBlock) and self.name == other.name and self.size == other.size

    def __lt__(self, other):
        if not isinstance(other, TempBlock):
            return NotImplemented
        return str(self) < str(other)

    def __le__(self, other):
        if not isinstance(other, TempBlock):
            return NotImplemented
        return str(self) <= str(other)

    def __gt__(self, other):
        if not isinstance(other, TempBlock):
            return NotImplemented
        return str(self) > str(other)

    def __ge__(self, other):
        if not isinstance(other, TempBlock):
            return NotImplemented
        return str(self) >= str(other)

    def __hash__(self):
        return self.__hash


class BlockPlace:
    __slots__ = ("__hash", "block", "index", "offset")

    def __init__(self, block: BlockValue, index: IndexValue = 0, offset: int = 0):
        self.block = block
        self.index = index
        self.offset = offset
        self.__hash = hash((block, index, offset))

    def __repr__(self):
        return f"BlockPlace(block={self.block!r}, index={self.index!r}, offset={self.offset!r})"

    def __str__(self):
        if isinstance(self.block, TempBlock) and self.block.size == 1 and self.index == 0 and self.offset == 0:
            return f"{self.block}"
        elif isinstance(self.index, int):
            return f"{self.block}[{self.index + self.offset}]"
        elif self.offset == 0:
            return f"{self.block}[{self.index}]"
        else:
            return f"{self.block}[{self.index} + {self.offset}]"

    def add_offset(self, offset: int) -> BlockPlace:
        return BlockPlace(self.block, self.index, self.offset + offset)

    def __eq__(self, other):
        return (
            isinstance(other, BlockPlace)
            and self.block == other.block
            and self.index == other.index
            and self.offset == other.offset
        )

    def __lt__(self, other):
        if not isinstance(other, BlockPlace):
            return NotImplemented
        return str(self) < str(other)

    def __le__(self, other):
        if not isinstance(other, BlockPlace):
            return NotImplemented
        return str(self) <= str(other)

    def __gt__(self, other):
        if not isinstance(other, BlockPlace):
            return NotImplemented
        return str(self) > str(other)

    def __ge__(self, other):
        if not isinstance(other, BlockPlace):
            return NotImplemented
        return str(self) >= str(other)

    def __hash__(self):
        return self.__hash

    def __iter__(self):
        yield self.block
        yield self.index
        yield self.offset


class SSAPlace:
    __slots__ = ("__hash", "name", "num")

    def __init__(self, name: str, num: int):
        self.name = name
        self.num = num
        self.__hash = hash((name, num))

    def __repr__(self):
        return f"SSAPlace(name={self.name!r}, num={self.num!r})"

    def __str__(self):
        return f"{self.name}.{self.num}"

    def __eq__(self, other):
        return isinstance(other, SSAPlace) and self.name == other.name and self.num == other.num

    def __lt__(self, other):
        if not isinstance(other, SSAPlace):
            return NotImplemented
        return str(self) < str(other)

    def __le__(self, other):
        if not isinstance(other, SSAPlace):
            return NotImplemented
        return str(self) <= str(other)

    def __gt__(self, other):
        if not isinstance(other, SSAPlace):
            return NotImplemented
        return str(self) > str(other)

    def __ge__(self, other):
        if not isinstance(other, SSAPlace):
            return NotImplemented
        return str(self) >= str(other)

    def __hash__(self):
        return self.__hash

    def __iter__(self):
        yield self.name
        yield self.num


PREALLOCATED_TEMP_PLACE_COUNT = 16384
preallocated_temp_block_places: list[BlockPlace] = [
    BlockPlace(TempBlock(f"v{i}", 1), 0) for i in range(PREALLOCATED_TEMP_PLACE_COUNT)
]



from __future__ import annotations

from collections.abc import Callable
from functools import wraps
from typing import overload


@overload
def meta_fn[T: Callable](fn: T) -> T: ...


@overload
def meta_fn[T: Callable](show_in_stack: bool) -> Callable[[T], T]: ...


def meta_fn(fn=None, *, show_in_stack: bool = True):
    """Marks a function as a meta function to be called directly without the AST visitor.

    This can also improve performance in some cases by avoiding the overhead of the AST visitor.
    """

    # noinspection PyShadowingNames
    def decorator(fn):
        from sonolus.backend import visitor
        from sonolus.backend.utils import get_function

        base_fn = fn
        while hasattr(base_fn, "__wrapped__"):
            base_fn = base_fn.__wrapped__

        function_name = getattr(base_fn, "__name__", "<unnamed>")
        module = getattr(base_fn, "__module__", None)
        if module is not None:
            qualified_name = f"{module}.{getattr(base_fn, '__qualname__', function_name)}"
        else:
            qualified_name = f"<unknown>.{getattr(base_fn, '__qualname__', function_name)}"
        source_file, node = get_function(base_fn)

        @wraps(fn)
        def wrapper(*args, **kwargs):
            if ctx():
                completion_timer = visitor.mark_start(qualified_name)
                debug_stack = ctx().callback_state.debug_stack
                try:
                    if show_in_stack:
                        debug_stack.append(f'File "{source_file}", line {node.lineno}, in {function_name}')
                    return fn(*args, **kwargs)
                finally:
                    if show_in_stack:
                        debug_stack.pop()
                    completion_timer()
            return fn(*args, **kwargs)

        wrapper._meta_fn_ = True
        return wrapper

    if fn is None:
        return decorator
    return decorator(fn)


# To indicate this was used for performance reasons rather than functionality
perf_meta_fn = meta_fn

from sonolus.script.internal.context import ctx



from dataclasses import dataclass, field
from enum import StrEnum


class UiMetric(StrEnum):
    """A metric to display on the UI."""

    ARCADE = "arcade"
    ARCADE_PERCENTAGE = "arcadePercentage"
    ACCURACY = "accuracy"
    ACCURACY_PERCENTAGE = "accuracyPercentage"
    LIFE = "life"
    PERFECT = "perfect"
    PERFECT_PERCENTAGE = "perfectPercentage"
    GREAT_GOOD_MISS = "greatGoodMiss"
    GREAT_GOOD_MISS_PERCENTAGE = "greatGoodMissPercentage"
    MISS = "miss"
    MISS_PERCENTAGE = "missPercentage"
    ERROR_HEATMAP = "errorHeatmap"


class UiJudgmentErrorStyle(StrEnum):
    """The style of the judgment error.

    The name of each member refers to what's used for positive (late) judgment errors.
    """

    NONE = "none"
    LATE = "late"
    EARLY = "early"  # Not really useful
    PLUS = "plus"
    MINUS = "minus"
    ARROW_UP = "arrowUp"
    ARROW_DOWN = "arrowDown"
    ARROW_LEFT = "arrowLeft"
    ARROW_RIGHT = "arrowRight"
    TRIANGLE_UP = "triangleUp"
    TRIANGLE_DOWN = "triangleDown"
    TRIANGLE_LEFT = "triangleLeft"
    TRIANGLE_RIGHT = "triangleRight"


class UiJudgmentErrorPlacement(StrEnum):
    """The placement of the judgment error."""

    LEFT = "left"
    RIGHT = "right"
    LEFT_RIGHT = "leftRight"
    TOP = "top"
    BOTTOM = "bottom"
    TOP_BOTTOM = "topBottom"
    CENTER = "center"


class EaseType(StrEnum):
    """The easing function to use."""

    LINEAR = "linear"
    NONE = "none"
    IN_SINE = "inSine"
    IN_QUAD = "inQuad"
    IN_CUBIC = "inCubic"
    IN_QUART = "inQuart"
    IN_QUINT = "inQuint"
    IN_EXPO = "inExpo"
    IN_CIRC = "inCirc"
    IN_BACK = "inBack"
    IN_ELASTIC = "inElastic"
    OUT_SINE = "outSine"
    OUT_QUAD = "outQuad"
    OUT_CUBIC = "outCubic"
    OUT_QUART = "outQuart"
    OUT_QUINT = "outQuint"
    OUT_EXPO = "outExpo"
    OUT_CIRC = "outCirc"
    OUT_BACK = "outBack"
    OUT_ELASTIC = "outElastic"
    IN_OUT_SINE = "inOutSine"
    IN_OUT_QUAD = "inOutQuad"
    IN_OUT_CUBIC = "inOutCubic"
    IN_OUT_QUART = "inOutQuart"
    IN_OUT_QUINT = "inOutQuint"
    IN_OUT_EXPO = "inOutExpo"
    IN_OUT_CIRC = "inOutCirc"
    IN_OUT_BACK = "inOutBack"
    IN_OUT_ELASTIC = "inOutElastic"
    OUT_IN_SINE = "outInSine"
    OUT_IN_QUAD = "outInQuad"
    OUT_IN_CUBIC = "outInCubic"
    OUT_IN_QUART = "outInQuart"
    OUT_IN_QUINT = "outInQuint"
    OUT_IN_EXPO = "outInExpo"
    OUT_IN_CIRC = "outInCirc"
    OUT_IN_BACK = "outInBack"
    OUT_IN_ELASTIC = "outInElastic"


@dataclass
class UiAnimationTween:
    """Tween animation configuration for UI elements.

    Args:
        start: The initial value.
        end: The final value.
        duration: The duration of the animation.
        ease: The easing function to use.
    """

    start: float
    end: float
    duration: float
    ease: EaseType

    def to_dict(self):
        return {
            "from": self.start,
            "to": self.end,
            "duration": self.duration,
            "ease": self.ease,
        }


@dataclass
class UiAnimation:
    """Animation configuration for UI elements.

    Args:
        scale: The animation applied to scale.
        alpha: The animation applied to alpha.
    """

    scale: UiAnimationTween
    alpha: UiAnimationTween

    def to_dict(self):
        return {
            "scale": self.scale.to_dict(),
            "alpha": self.alpha.to_dict(),
        }


@dataclass
class UiVisibility:
    """Visibility configuration for UI elements.

    Args:
        scale: The scale of the element.
        alpha: The alpha of the element.
    """

    scale: float = 1.0
    alpha: float = 1.0

    def to_dict(self):
        return {
            "scale": self.scale,
            "alpha": self.alpha,
        }


@dataclass
class UiConfig:
    """Configuration for UI elements.

    Args:
        scope: The scope of the configuration.
        primary_metric: The primary metric to display.
        secondary_metric: The secondary metric to display.
        menu_visibility: The visibility configuration for the menu.
        judgment_visibility: The visibility configuration for judgments.
        combo_visibility: The visibility configuration for the combo.
        primary_metric_visibility: The visibility configuration for the primary metric.
        secondary_metric_visibility: The visibility configuration for the secondary metric.
        progress_visibility: The visibility configuration for progress.
        tutorial_navigation_visibility: The visibility configuration for tutorial navigation.
        tutorial_instruction_visibility: The visibility configuration for tutorial instructions.
        judgment_animation: The animation configuration for judgments.
        combo_animation: The animation configuration for the combo.
        judgment_error_style: The style of the judgment error.
        judgment_error_placement: The placement of the judgment error.
        judgment_error_min: The minimum judgment error.
    """

    scope: str | None = None
    primary_metric: UiMetric = UiMetric.ARCADE
    secondary_metric: UiMetric = UiMetric.LIFE
    menu_visibility: UiVisibility = field(default_factory=UiVisibility)
    judgment_visibility: UiVisibility = field(default_factory=UiVisibility)
    combo_visibility: UiVisibility = field(default_factory=UiVisibility)
    primary_metric_visibility: UiVisibility = field(default_factory=UiVisibility)
    secondary_metric_visibility: UiVisibility = field(default_factory=UiVisibility)
    progress_visibility: UiVisibility = field(default_factory=UiVisibility)
    tutorial_navigation_visibility: UiVisibility = field(default_factory=UiVisibility)
    tutorial_instruction_visibility: UiVisibility = field(default_factory=UiVisibility)
    judgment_animation: UiAnimation = field(
        default_factory=lambda: UiAnimation(
            scale=UiAnimationTween(0, 1, 0.1, EaseType.OUT_CUBIC), alpha=UiAnimationTween(1, 0, 0.3, EaseType.NONE)
        )
    )
    combo_animation: UiAnimation = field(
        default_factory=lambda: UiAnimation(
            scale=UiAnimationTween(1.2, 1, 0.2, EaseType.IN_CUBIC), alpha=UiAnimationTween(1, 1, 0, EaseType.NONE)
        )
    )
    judgment_error_style: UiJudgmentErrorStyle = UiJudgmentErrorStyle.LATE
    judgment_error_placement: UiJudgmentErrorPlacement = UiJudgmentErrorPlacement.TOP
    judgment_error_min: float = 0.0

    def to_dict(self):
        data = {
            "primaryMetric": self.primary_metric,
            "secondaryMetric": self.secondary_metric,
            "menuVisibility": self.menu_visibility.to_dict(),
            "judgmentVisibility": self.judgment_visibility.to_dict(),
            "comboVisibility": self.combo_visibility.to_dict(),
            "primaryMetricVisibility": self.primary_metric_visibility.to_dict(),
            "secondaryMetricVisibility": self.secondary_metric_visibility.to_dict(),
            "progressVisibility": self.progress_visibility.to_dict(),
            "tutorialNavigationVisibility": self.tutorial_navigation_visibility.to_dict(),
            "tutorialInstructionVisibility": self.tutorial_instruction_visibility.to_dict(),
            "judgmentAnimation": self.judgment_animation.to_dict(),
            "comboAnimation": self.combo_animation.to_dict(),
            "judgmentErrorStyle": self.judgment_error_style,
            "judgmentErrorPlacement": self.judgment_error_placement,
            "judgmentErrorMin": self.judgment_error_min,
        }
        if self.scope is not None:
            data["scope"] = self.scope
        return data



from __future__ import annotations

import inspect
from abc import abstractmethod, ABCMeta
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from enum import Enum, IntEnum, StrEnum
from types import FunctionType
from typing import Annotated, Any, ClassVar, NamedTuple, Self, TypedDict, get_origin

from sonolus.backend.ir import IRConst, IRExpr, IRInstr, IRPureInstr, IRStmt
from sonolus.backend.mode import Mode
from sonolus.backend.ops import Op
from sonolus.backend.visitor import compile_and_call
from sonolus.script.bucket import Bucket, Judgment
from sonolus.script.debug import runtime_checks_enabled, static_error
from sonolus.script.internal.callbacks import PLAY_CALLBACKS, PREVIEW_CALLBACKS, WATCH_ARCHETYPE_CALLBACKS, CallbackInfo
from sonolus.script.internal.context import ctx
from sonolus.script.internal.descriptor import SonolusDescriptor
from sonolus.script.internal.generic import validate_concrete_type
from sonolus.script.internal.impl import validate_value
from sonolus.script.internal.introspection import get_field_specifiers
from sonolus.script.internal.meta_fn import meta_fn
from sonolus.script.internal.native import native_call
from sonolus.script.internal.value import BackingValue, DataValue, Value
from sonolus.script.num import Num
from sonolus.script.pointer import _backing_deref, _deref
from sonolus.script.record import Record
from sonolus.script.timing import TimescaleEase
from sonolus.script.values import zeros

_ENTITY_MEMORY_SIZE = 64
_ENTITY_DATA_SIZE = 32
_ENTITY_SHARED_MEMORY_SIZE = 32


class _StorageType(Enum):
    IMPORTED = "imported"
    EXPORTED = "exported"
    MEMORY = "memory"
    SHARED = "shared_memory"


@dataclass
class _ArchetypeFieldInfo:
    name: str | None
    storage: _StorageType
    default: Value | None = None


class _ExportBackingValue(BackingValue):
    def __init__(self, index: IRExpr):
        self.index = index

    def read(self) -> IRExpr:
        raise NotImplementedError("Exported fields are write-only")

    def write(self, value: IRExpr) -> IRStmt:
        return IRInstr(Op.ExportValue, [self.index, value])


class _ArchetypeField(SonolusDescriptor):
    def __init__(
        self,
        name: str,
        data_name: str,
        storage: _StorageType,
        offset: int,
        type_: type[Value],
        default: Value | None = None,
    ):
        self.name = name
        self.data_name = data_name  # name used in level data
        self.storage = storage
        self.offset = offset
        self.type = type_
        self.default = default

    def __get__(self, instance: _BaseArchetype, owner):
        if instance is None:
            return self
        result = None
        match self.storage:
            case _StorageType.IMPORTED:
                match instance._data_:
                    case _ArchetypeSelfData():
                        result = _deref(ctx().blocks.EntityData, self.offset, self.type)
                    case _ArchetypeReferenceData(index=index):
                        result = _deref(
                            ctx().blocks.EntityDataArray,
                            Num._accept_(self.offset) + index * _ENTITY_DATA_SIZE,
                            self.type,
                        )
                    case _ArchetypeLevelData(values=values):
                        result = values[self.name]
            case _StorageType.EXPORTED:
                match instance._data_:
                    case _ArchetypeSelfData():

                        def backing_source(i: IRExpr):
                            return _ExportBackingValue(IRPureInstr(Op.Add, [i, IRConst(self.offset)]))

                        result = _backing_deref(
                            backing_source,
                            self.type,
                        )
                    case _ArchetypeReferenceData():
                        raise RuntimeError("Exported fields of other entities are not accessible")
                    case _ArchetypeLevelData():
                        raise RuntimeError("Exported fields are not available in level data")
            case _StorageType.MEMORY:
                match instance._data_:
                    case _ArchetypeSelfData():
                        result = _deref(ctx().blocks.EntityMemory, self.offset, self.type)
                    case _ArchetypeReferenceData():
                        raise RuntimeError("Entity memory of other entities is not accessible")
                    case _ArchetypeLevelData():
                        raise RuntimeError("Entity memory is not available in level data")
            case _StorageType.SHARED:
                match instance._data_:
                    case _ArchetypeSelfData():
                        result = _deref(ctx().blocks.EntitySharedMemory, self.offset, self.type)
                    case _ArchetypeReferenceData(index=index):
                        result = _deref(
                            ctx().blocks.EntitySharedMemoryArray,
                            Num._accept_(self.offset) + index * _ENTITY_SHARED_MEMORY_SIZE,
                            self.type,
                        )
                    case _ArchetypeLevelData():
                        raise RuntimeError("Entity shared memory is not available in level data")
        if result is None:
            raise RuntimeError("Invalid storage type")
        if ctx():
            return result._get_readonly_()
        else:
            return result._as_py_()  # type: ignore

    def __set__(self, instance: _BaseArchetype, value):
        if instance is None:
            raise RuntimeError("Cannot set field on class")
        if not self.type._accepts_(value):
            raise TypeError(f"Expected {self.type}, got {type(value)}")
        target = None
        match self.storage:
            case _StorageType.IMPORTED:
                match instance._data_:
                    case _ArchetypeSelfData():
                        target = _deref(ctx().blocks.EntityData, self.offset, self.type)
                    case _ArchetypeReferenceData(index=index):
                        target = _deref(
                            ctx().blocks.EntityDataArray,
                            Num._accept_(self.offset) + index * _ENTITY_DATA_SIZE,
                            self.type,
                        )
                    case _ArchetypeLevelData(values=values):
                        target = values[self.name]
            case _StorageType.EXPORTED:
                match instance._data_:
                    case _ArchetypeSelfData():
                        if not isinstance(value, self.type):
                            raise TypeError(f"Expected {self.type}, got {type(value)}")
                        for k, v in value._to_flat_dict_(self.data_name).items():
                            index = instance._exported_keys_[k]
                            ctx().add_statements(IRInstr(Op.ExportValue, [IRConst(index), Num(v).ir()]))
                        return
                    case _ArchetypeReferenceData():
                        raise RuntimeError("Exported fields of other entities are not accessible")
                    case _ArchetypeLevelData():
                        raise RuntimeError("Exported fields are not available in level data")
            case _StorageType.MEMORY:
                match instance._data_:
                    case _ArchetypeSelfData():
                        target = _deref(ctx().blocks.EntityMemory, self.offset, self.type)
                    case _ArchetypeReferenceData():
                        raise RuntimeError("Entity memory of other entities is not accessible")
                    case _ArchetypeLevelData():
                        raise RuntimeError("Entity memory is not available in level data")
            case _StorageType.SHARED:
                match instance._data_:
                    case _ArchetypeSelfData():
                        target = _deref(ctx().blocks.EntitySharedMemory, self.offset, self.type)
                    case _ArchetypeReferenceData(index=index):
                        target = _deref(
                            ctx().blocks.EntitySharedMemoryArray,
                            Num._accept_(self.offset) + index * _ENTITY_SHARED_MEMORY_SIZE,
                            self.type,
                        )
                    case _ArchetypeLevelData():
                        raise RuntimeError("Entity shared memory is not available in level data")
        if target is None:
            raise RuntimeError("Invalid storage type")
        value = self.type._accept_(value)
        if self.type._is_value_type_():
            target._set_(value)
        else:
            target._copy_from_(value)


class _NameDescriptor(SonolusDescriptor):
    def __init__(self, name: str):
        self.name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self.name
        elif ctx():
            raise RuntimeError("Cannot access archetype name in from self in a callback, use ArchetypeClass.name")
        else:
            return self.name

    def __set__(self, instance, value):
        raise AttributeError("Archetype name is read-only and cannot be set")


class _IsScoredDescriptor(SonolusDescriptor):
    def __init__(self, value: bool):
        self.value = value

    def __get__(self, instance, owner):
        if instance is None:
            return self.value
        elif ctx():
            return ctx().mode_state.is_scored_by_archetype_id.get_unchecked(instance.id)
        else:
            return self.value

    def __set__(self, instance, value):
        raise AttributeError("is_scored is read-only and cannot be set")


class _IdDescriptor(SonolusDescriptor):
    def __get__(self, instance, owner):
        if not ctx():
            raise RuntimeError("Archetype id is only available during compilation")
        if instance is None:
            result = ctx().mode_state.archetypes.get(owner)
            if result is None or owner in ctx().mode_state.compile_time_only_archetypes:
                raise RuntimeError("Archetype is not registered")
            return result
        else:
            return instance._info.archetype_id

    def __set__(self, instance, value):
        raise AttributeError("Archetype id is read-only and cannot be set")


class _KeyDescriptor(SonolusDescriptor):
    def __init__(self, value: int | float):
        self.value = value

    def __get__(self, instance, owner):
        if instance is not None and ctx():
            return ctx().mode_state.keys_by_archetype_id.get_unchecked(instance.id)
        else:
            return self.value

    def __set__(self, instance, value):
        raise AttributeError("Archetype key is read-only and cannot be set")


class _ArchetypeLifeDescriptor(SonolusDescriptor):
    def __get__(self, instance, owner):
        if not ctx():
            raise RuntimeError("Archetype life is only available during compilation")
        if ctx().mode_state.mode not in {Mode.PLAY, Mode.WATCH}:
            raise RuntimeError(f"Archetype life is not available in mode '{ctx().mode_state.mode.value}'")
        if instance is not None:
            return _deref(ctx().blocks.ArchetypeLife, instance.id * LifeInfo._size_(), LifeInfo)
        else:
            return _deref(ctx().blocks.ArchetypeLife, owner.id * LifeInfo._size_(), LifeInfo)

    def __set__(self, instance, value):
        raise AttributeError("Archetype life is read-only and cannot be set")


class _EntityLifeDescriptor(SonolusDescriptor):
    def __get__(self, instance, owner):
        if not ctx():
            raise RuntimeError("Entity life is only available during compilation")
        if ctx().mode_state.mode not in {Mode.PLAY, Mode.WATCH}:
            raise RuntimeError(f"Entity life is not available in mode '{ctx().mode_state.mode.value}'")
        if instance is None:
            raise RuntimeError("Entity life can only be accessed from an instance")
        match instance._data_:
            case _ArchetypeSelfData():
                return _deref(ctx().blocks.EntityLife, 0, LifeInfo)
            case _ArchetypeReferenceData(index=index):
                return _deref(ctx().blocks.EntityLifeArray, index * LifeInfo._size_(), LifeInfo)
            case _:
                raise RuntimeError("Entity life is not available in level data")

    def __set__(self, instance, value):
        raise AttributeError("Entity life is read-only and cannot be set")


class _ArchetypeScoreMultiplierMetaDescriptor(SonolusDescriptor):
    def __get__(self, instance, owner):
        # instance is the class (e.g., MyArchetype)
        # owner is the metaclass (_BaseArchetypeMeta)
        if instance is None:
            return self
        if not ctx():
            raise RuntimeError("Archetype score multiplier is only available during compilation")
        if ctx().mode_state.mode not in {Mode.PLAY, Mode.WATCH}:
            raise RuntimeError(f"Archetype score multiplier is not available in mode '{ctx().mode_state.mode.value}'")
        return _deref(ctx().blocks.ArchetypeScore, instance.id, Num)

    def __set__(self, instance, value):
        # instance is the class
        if not ctx():
            raise RuntimeError("Archetype score multiplier is only available during compilation")
        if ctx().mode_state.mode not in {Mode.PLAY, Mode.WATCH}:
            raise RuntimeError(f"Archetype score multiplier is not available in mode '{ctx().mode_state.mode.value}'")
        target = _deref(ctx().blocks.ArchetypeScore, instance.id, Num)
        target._set_(Num._accept_(value))


class _ArchetypeScoreMultiplierDescriptor(SonolusDescriptor):
    def __get__(self, instance, owner):
        if not ctx():
            raise RuntimeError("Archetype score multiplier is only available during compilation")
        if ctx().mode_state.mode not in {Mode.PLAY, Mode.WATCH}:
            raise RuntimeError(f"Archetype score multiplier is not available in mode '{ctx().mode_state.mode.value}'")
        if instance is not None:
            return _deref(ctx().blocks.ArchetypeScore, instance.id, Num)
        else:
            return _deref(ctx().blocks.ArchetypeScore, owner.id, Num)

    def __set__(self, instance, value):
        # Handle instance writes
        if instance is None:
            raise RuntimeError("Cannot set archetype score multiplier on None instance")
        if not ctx():
            raise RuntimeError("Archetype score multiplier is only available during compilation")
        if ctx().mode_state.mode not in {Mode.PLAY, Mode.WATCH}:
            raise RuntimeError(f"Archetype score multiplier is not available in mode '{ctx().mode_state.mode.value}'")
        target = _deref(ctx().blocks.ArchetypeScore, instance.id, Num)
        target._set_(Num._accept_(value))


class _EntityScoreMultiplierMetaDescriptor(SonolusDescriptor):
    def __get__(self, instance, owner):
        # instance is the class (e.g., MyArchetype)
        # owner is the metaclass (_BaseArchetypeMeta)
        if instance is None:
            return self
        raise RuntimeError("Entity score multiplier can only be accessed from an instance")

    def __set__(self, instance, value):
        # instance is the class
        raise RuntimeError("Entity score multiplier can only be set on an instance, not on the class")


class _EntityScoreMultiplierDescriptor(SonolusDescriptor):
    def __get__(self, instance, owner):
        if not ctx():
            raise RuntimeError("Entity score multiplier is only available during compilation")
        if ctx().mode_state.mode not in {Mode.PLAY, Mode.WATCH}:
            raise RuntimeError(f"Entity score multiplier is not available in mode '{ctx().mode_state.mode.value}'")
        if instance is None:
            raise RuntimeError("Entity score multiplier can only be accessed from an instance")
        match instance._data_:
            case _ArchetypeSelfData():
                return _deref(ctx().blocks.EntityScore, 0, Num)
            case _ArchetypeReferenceData(index=index):
                return _deref(ctx().blocks.EntityScoreArray, index, Num)
            case _:
                raise RuntimeError("Entity score multiplier is not available in level data")

    def __set__(self, instance, value):
        # Handle instance writes
        if instance is None:
            raise RuntimeError("Entity score multiplier can only be set on an instance")
        if not ctx():
            raise RuntimeError("Entity score multiplier is only available during compilation")
        if ctx().mode_state.mode not in {Mode.PLAY, Mode.WATCH}:
            raise RuntimeError(f"Entity score multiplier is not available in mode '{ctx().mode_state.mode.value}'")
        target = None
        match instance._data_:
            case _ArchetypeSelfData():
                target = _deref(ctx().blocks.EntityScore, 0, Num)
            case _ArchetypeReferenceData(index=index):
                target = _deref(ctx().blocks.EntityScoreArray, index, Num)
            case _:
                raise RuntimeError("Entity score multiplier is not available in level data")
        target._set_(Num._accept_(value))


def imported(*, name: str | None = None, default: int | float | None = None) -> Any:
    """Declare a field as imported.

    Imported fields may be loaded from the level.

    In watch mode, data may also be loaded from a corresponding exported field in play mode.

    Imported fields may only be updated in the [`preprocess`][sonolus.script.archetype.PlayArchetype.preprocess]
    callback, and are read-only in other callbacks.

    Usage:
        ```python
        class MyArchetype(PlayArchetype):
            field: int = imported()
            field_with_explicit_name: int = imported(name="field_name")
            field_with_default: int = imported(default=0)
        ```
    """
    validated_default = None
    if default is not None:
        validated_default = validate_value(default)
    return _ArchetypeFieldInfo(name, _StorageType.IMPORTED, validated_default)


def entity_data() -> Any:
    """Declare a field as entity data.

    Entity data is accessible from other entities, but may only be updated in the
    [`preprocess`][sonolus.script.archetype.PlayArchetype.preprocess] callback
    and is read-only in other callbacks.

    It functions like [`imported`][sonolus.script.archetype.imported] and shares the same underlying storage,
    except that it is not loaded from a level.

    Usage:
        ```python
        class MyArchetype(PlayArchetype):
            field: int = entity_data()
        ```
    """
    return _ArchetypeFieldInfo(None, _StorageType.IMPORTED)


def exported(*, name: str | None = None) -> Any:
    """Declare a field as exported.

    This is only usable in play mode to export data to be loaded in watch mode.

    Exported fields are write-only.

    Usage:
        ```python
        class MyArchetype(PlayArchetype):
            field: int = exported()
            field_with_explicit_name: int = exported(name="#FIELD")
        ```
    """
    return _ArchetypeFieldInfo(name, _StorageType.EXPORTED)


def entity_memory() -> Any:
    """Declare a field as entity memory.

    Entity memory is private to the entity and is not accessible from other entities. It may be read or updated in any
    callback associated with the entity.

    Entity memory fields may also be set when an entity is spawned using the
    [`spawn()`][sonolus.script.archetype.PlayArchetype.spawn] method.

    Usage:
        ```python
        class MyArchetype(PlayArchetype):
            field: int = entity_memory()

        ```
    """
    return _ArchetypeFieldInfo(None, _StorageType.MEMORY)


def shared_memory() -> Any:
    """Declare a field as shared memory.

    Shared memory is accessible from other entities.

    Shared memory may be read in any callback, but may only be updated by sequential callbacks
    ([`preprocess`][sonolus.script.archetype.PlayArchetype.preprocess],
    [`update_sequential`][sonolus.script.archetype.PlayArchetype.update_sequential],
    and [`touch`][sonolus.script.archetype.PlayArchetype.touch]).

    Usage:
        ```python
        class MyArchetype(PlayArchetype):
            field: int = shared_memory()
        ```
    """
    return _ArchetypeFieldInfo(None, _StorageType.SHARED)


_annotation_defaults: dict[Callable, _ArchetypeFieldInfo] = {
    imported: imported(),
    exported: exported(),
    entity_memory: entity_memory(),
    shared_memory: shared_memory(),
}


def callback[T: Callable](*, order: int = 0) -> Callable[[T], T]:
    """Annotate a callback with its order.

    Callbacks are executed from lowest to highest order. By default, callbacks have an order of 0.

    Usage:
        ```python
        class MyArchetype(PlayArchetype):
            @callback(order=1)
            def update_sequential(self):
                pass
        ```

    Args:
        order: The order of the callback. Lower values are executed first.
    """

    def decorator(func: T) -> T:
        func._callback_order_ = order  # type: ignore
        return func

    return decorator


class _ArchetypeSelfData:
    pass


class _ArchetypeReferenceData:
    index: int

    def __init__(self, index: int):
        self.index = index


class _ArchetypeLevelData:
    values: dict[str, Value]

    def __init__(self, values: dict[str, Value]):
        self.values = values


type _ArchetypeData = _ArchetypeSelfData | _ArchetypeReferenceData | _ArchetypeLevelData


class ArchetypeSchema(TypedDict):
    name: str
    fields: list[str]


class ImportInfo(NamedTuple):
    index: int
    default: int | float | None


RESERVED_ARCHETYPE_FIELD_NAMES = {
    "id",
    "key",
    "name",
    "life",
    "archetype_life",
    "entity_life",
    "archetype_score_multiplier",
    "entity_score_multiplier",
    "is_scored",
    "_key_",
    "_is_scored_",
    "_derived_base_",
    "_is_derived_",
    "_default_callbacks_",
    "_callbacks_",
    "_field_init_done",
    "_is_concrete_archetype_",
    "_abc_impl",
}

ALLOWED_RESERVED_ARCHETYPE_FIELD_NAME_OVERRIDES = {
    "key",
    "name",
    "is_scored",
}


class _BaseArchetypeMeta(ABCMeta):
    archetype_score_multiplier = _ArchetypeScoreMultiplierMetaDescriptor()
    entity_score_multiplier = _EntityScoreMultiplierMetaDescriptor()

    def __new__(mcs, name, bases, namespace, **kwargs):
        module = namespace.get("__module__", "")
        is_derived = namespace.get("_is_derived_", False)
        if module != "sonolus.script.archetype" and not is_derived:
            for field_name in namespace:
                if (
                    field_name in RESERVED_ARCHETYPE_FIELD_NAMES
                    and field_name not in ALLOWED_RESERVED_ARCHETYPE_FIELD_NAME_OVERRIDES
                ):
                    raise TypeError(f"Field '{field_name}' in {name} overrides a reserved archetype field name.")

        return super().__new__(mcs, name, bases, namespace, **kwargs)


class _BaseArchetype(metaclass=_BaseArchetypeMeta):
    _is_comptime_value_ = True

    _removable_prefix: ClassVar[str] = ""

    _supported_callbacks_: ClassVar[dict[str, CallbackInfo]]
    _default_callbacks_: ClassVar[set[Callable]]

    _imported_fields_: ClassVar[dict[str, _ArchetypeField]]
    _exported_fields_: ClassVar[dict[str, _ArchetypeField]]
    _memory_fields_: ClassVar[dict[str, _ArchetypeField]]
    _shared_memory_fields_: ClassVar[dict[str, _ArchetypeField]]

    _imported_keys_: ClassVar[dict[str, ImportInfo]]
    _exported_keys_: ClassVar[dict[str, int]]
    _callbacks_: ClassVar[dict[str, Callable]]
    _data_constructor_signature_: ClassVar[inspect.Signature]
    _spawn_signature_: ClassVar[inspect.Signature]

    _data_: _ArchetypeData

    archetype_score_multiplier = _ArchetypeScoreMultiplierDescriptor()
    entity_score_multiplier = _EntityScoreMultiplierDescriptor()

    id: int = 0
    """The id of the archetype or entity.

    If accessed on an entity, always returns the runtime archetype id of the entity, even if it doesn't match the type
    that was used to access it.

    E.g. if an entity of archetype `A` is accessed via [`EntityRef[B]`][sonolus.script.archetype.EntityRef], the id will
    still be the id of `A`.
    """

    key: int | float = -1
    """An optional key for the archetype.

    May be useful to identify an archetype in an inheritance hierarchy without needing to check id.

    If accessed on an entity, always returns the runtime key of the entity, even if it doesn't match the type
    that was used to access it.

    E.g. if an entity of archetype `A` is accessed via [`EntityRef[B]`][sonolus.script.archetype.EntityRef], the key
    will still be the key of `A`.
    """

    name: ClassVar[str | None] = None
    """The name of the archetype.

    If not set, the name will be the class name.

    The name is used in level data.
    """

    is_scored: ClassVar[bool] = False

    def __init__(self, *args, **kwargs):
        self._init_fields()
        if ctx():
            raise RuntimeError("The Archetype constructor is only for defining level data")
        bound = self._data_constructor_signature_.bind_partial(*args, **kwargs)
        bound.apply_defaults()
        values = {
            field.name: field.type._accept_(bound.arguments.get(field.name) or zeros(field.type))._get_()
            for field in self._imported_fields_.values()
        }
        self._data_ = _ArchetypeLevelData(values=values)

    @classmethod
    def _new(cls):
        cls._init_fields()
        return object.__new__(cls)

    @classmethod
    def _for_compilation(cls):
        cls._init_fields()
        result = cls._new()
        result._data_ = _ArchetypeSelfData()
        return result

    @classmethod
    @meta_fn
    def _compile_time_id(cls):
        if not ctx():
            raise RuntimeError("Archetype id is only available during compilation")
        result = ctx().mode_state.archetypes.get(cls)
        if result is None:
            raise RuntimeError("Archetype is not registered")
        return result

    @classmethod
    @meta_fn
    def at(cls, index: int, check: bool = True) -> Self:
        """Access the entity of this archetype at the given index.

        Args:
            index: The index of the entity to reference.
            check: If true, raises an error if the entity at the index is not of this archetype or of a subclass of
                   this archetype. If false, no validation is performed.

        Returns:
            The entity at the given index.
        """
        if not ctx():
            raise RuntimeError("Archetype.at is only available during compilation")
        compile_and_call(cls._check_is_at, index, check=check)
        result = cls._new()
        result._data_ = _ArchetypeReferenceData(index=Num._accept_(index))
        return result

    @classmethod
    def is_at(cls, index: int, strict: bool = False) -> bool:
        """Return whether the entity at the given index is of this archetype.

        Args:
            index: The index of the entity to check.
            strict: If true, only returns true if the entity is exactly of this archetype. If false, also returns true
                    if the entity is of a subclass of this archetype.

        Returns:
            Whether the entity at the given index is of this archetype.
        """
        if strict:
            return index >= 0 and cls._compile_time_id() == entity_info_at(index).archetype_id
        else:
            mro_ids = cls._get_mro_id_array(entity_info_at(index).archetype_id)
            return index >= 0 and cls._compile_time_id() in mro_ids

    @classmethod
    def _check_is_at(cls, index: int, check: bool):
        if not check or not runtime_checks_enabled():
            return
        assert index >= 0, "Entity index must be non-negative"
        assert cls._compile_time_id() in cls._get_mro_id_array(entity_info_at(index).archetype_id), (
            "Entity at index is not of the expected archetype"
        )

    @staticmethod
    @meta_fn
    def _get_mro_id_array(archetype_id: int) -> Sequence[int]:
        if not ctx():
            raise RuntimeError("Archetype._get_mro_id_array is only available during compilation")
        return ctx().get_archetype_mro_id_array(archetype_id)

    @classmethod
    @meta_fn
    def spawn(cls, **kwargs: Any) -> None:
        """Spawn an entity of this archetype, injecting the given values into entity memory.

        Usage:
            ```python
            class MyArchetype(PlayArchetype):
                field: int = entity_memory()

            def f():
                MyArchetype.spawn(field=123)
            ```

        Args:
            **kwargs: Entity memory values to inject by field name as defined in the Archetype.
        """
        cls._init_fields()
        if not ctx():
            raise RuntimeError("Spawn is only allowed within a callback")
        archetype_id = cls.id
        bound = cls._spawn_signature_.bind_partial(**kwargs)
        bound.apply_defaults()
        data = []
        for field in cls._memory_fields_.values():
            data.extend(
                field.type._accept_(
                    bound.arguments[field.name] if field.name in bound.arguments else zeros(field.type)
                )._to_list_()
            )
        native_call(Op.Spawn, archetype_id, *(Num(x) for x in data))

    @classmethod
    def schema(cls) -> ArchetypeSchema:
        cls._init_fields()
        return {"name": cls.name or "unnamed", "fields": list(cls._imported_fields_)}

    def _level_data_entries(self, level_refs: dict[Any, str] | None = None):
        self._init_fields()
        if not isinstance(self._data_, _ArchetypeLevelData):
            raise RuntimeError("Entity is not level data")
        entries = []
        for name, value in self._data_.values.items():
            field_info = self._imported_fields_.get(name)
            for k, v in value._to_flat_dict_(field_info.data_name, level_refs).items():
                if isinstance(v, str):
                    entries.append({"name": k, "ref": v})
                else:
                    entries.append({"name": k, "value": v})
        return entries

    def __init_subclass__(cls, **kwargs):
        if cls.__module__ == _BaseArchetype.__module__ and not getattr(cls, "_is_derived_", False):
            if cls._supported_callbacks_ is None:
                raise TypeError("Cannot directly subclass Archetype, use the Archetype subclass for your mode")
            cls._default_callbacks_ = {getattr(cls, cb_info.py_name) for cb_info in cls._supported_callbacks_.values()}
            return
        if cls.name is None or cls.name in {getattr(mro_entry, "name", None) for mro_entry in cls.mro()[1:]}:
            cls.name = cls.__name__.removeprefix(cls._removable_prefix)
        cls._callbacks_ = {}
        for name in cls._supported_callbacks_:
            for mro_entry in cls.mro():
                if name in mro_entry.__dict__:
                    cb = mro_entry.__dict__[name]
                    if cb not in cls._default_callbacks_:
                        cls._callbacks_[name] = cb
                        break
        cls._field_init_done = False
        cls._is_concrete_archetype_ = True
        cls.id = _IdDescriptor()
        cls._key_ = cls.key
        cls.key = _KeyDescriptor(cls.key)
        cls.name = _NameDescriptor(cls.name)
        cls._is_scored_ = cls.is_scored
        cls.is_scored = _IsScoredDescriptor(cls.is_scored)
        cls.life = _ArchetypeLifeDescriptor()
        cls.archetype_life = _ArchetypeLifeDescriptor()
        cls.entity_life = _EntityLifeDescriptor()

    @classmethod
    def _init_fields(cls):
        if cls._field_init_done:
            return
        for mro_entry in cls.mro()[1:]:
            if hasattr(mro_entry, "_field_init_done"):
                mro_entry._init_fields()
        if sum(issubclass(base, _BaseArchetype) for base in cls.__bases__) > 1:
            raise TypeError("Multiple inheritance of Archetypes is not supported")
        mro_from_archetype_parents = set(
            type("Dummy", tuple(base for base in cls.__bases__ if issubclass(base, _BaseArchetype)), {}).mro()
        )
        # Archetype parents would have already initialized relevant fields, so only consider the current class
        # and mixins that were not already included via an archetype parent
        mro_excluding_archetype_parents = [entry for entry in cls.mro() if entry not in mro_from_archetype_parents]
        try:
            field_specifiers = get_field_specifiers(
                cls,
                skip=RESERVED_ARCHETYPE_FIELD_NAMES,
                included_classes=mro_excluding_archetype_parents,
            ).items()
        except Exception as e:
            raise TypeError(f"Error while processing fields of {cls.__name__}: {e}") from e
        if not hasattr(cls, "_imported_fields_"):
            cls._imported_fields_ = {}
        else:
            cls._imported_fields_ = {**cls._imported_fields_}
        if not hasattr(cls, "_exported_fields_"):
            cls._exported_fields_ = {}
        else:
            cls._exported_fields_ = {**cls._exported_fields_}
        if not hasattr(cls, "_memory_fields_"):
            cls._memory_fields_ = {}
        else:
            cls._memory_fields_ = {**cls._memory_fields_}
        if not hasattr(cls, "_shared_memory_fields_"):
            cls._shared_memory_fields_ = {}
        else:
            cls._shared_memory_fields_ = {**cls._shared_memory_fields_}
        imported_offset = sum(field.type._size_() for field in cls._imported_fields_.values())
        exported_offset = sum(field.type._size_() for field in cls._exported_fields_.values())
        memory_offset = sum(field.type._size_() for field in cls._memory_fields_.values())
        shared_memory_offset = sum(field.type._size_() for field in cls._shared_memory_fields_.values())
        for name, value in field_specifiers:
            if value is ClassVar or get_origin(value) is ClassVar:
                continue
            if get_origin(value) is not Annotated:
                raise TypeError(
                    "Archetype fields must be annotated using imported, exported, entity_memory, or shared_memory"
                )
            field_info = None
            for metadata in value.__metadata__:
                if isinstance(metadata, FunctionType):
                    metadata = _annotation_defaults.get(metadata, metadata)
                if isinstance(metadata, _ArchetypeFieldInfo):
                    if field_info is not None:
                        if field_info.storage == metadata.storage and field_info.name is None:
                            field_info = metadata
                        elif field_info.storage == metadata.storage and (
                            metadata.name is None or field_info.name == metadata.name
                        ):
                            pass
                        else:
                            raise TypeError(
                                f"Unexpected multiple annotations for field '{name}' of {cls.__name__}, "
                                f"expected exactly one of imported, exported, entity_memory, or shared_memory"
                            )
                    else:
                        field_info = metadata
            if field_info is None:
                raise TypeError(
                    f"Missing annotation for '{name}' of {cls.__name__}, "
                    f"expected exactly one of imported, exported, entity_memory, or shared_memory"
                )
            if (
                name in cls._imported_fields_
                or name in cls._exported_fields_
                or name in cls._memory_fields_
                or name in cls._shared_memory_fields_
            ):
                raise ValueError(f"Field '{name}' is already defined in a superclass")
            try:
                field_type = validate_concrete_type(value.__args__[0])
            except Exception as e:
                raise TypeError(f"Error in field '{name}' of {cls.__name__}: {e}") from e
            match field_info.storage:
                case _StorageType.IMPORTED:
                    cls._imported_fields_[name] = _ArchetypeField(
                        name,
                        field_info.name or name,
                        field_info.storage,
                        imported_offset,
                        field_type,
                        field_info.default,
                    )
                    imported_offset += field_type._size_()
                    if imported_offset > _ENTITY_DATA_SIZE:
                        raise ValueError("Imported fields exceed entity data size")
                    setattr(cls, name, cls._imported_fields_[name])
                case _StorageType.EXPORTED:
                    cls._exported_fields_[name] = _ArchetypeField(
                        name, field_info.name or name, field_info.storage, exported_offset, field_type
                    )
                    exported_offset += field_type._size_()
                    if exported_offset > _ENTITY_DATA_SIZE:
                        raise ValueError("Exported fields exceed entity data size")
                    setattr(cls, name, cls._exported_fields_[name])
                case _StorageType.MEMORY:
                    cls._memory_fields_[name] = _ArchetypeField(
                        name, field_info.name or name, field_info.storage, memory_offset, field_type
                    )
                    memory_offset += field_type._size_()
                    if memory_offset > _ENTITY_MEMORY_SIZE:
                        raise ValueError("Memory fields exceed entity memory size")
                    setattr(cls, name, cls._memory_fields_[name])
                case _StorageType.SHARED:
                    cls._shared_memory_fields_[name] = _ArchetypeField(
                        name, field_info.name or name, field_info.storage, shared_memory_offset, field_type
                    )
                    shared_memory_offset += field_type._size_()
                    if shared_memory_offset > _ENTITY_SHARED_MEMORY_SIZE:
                        raise ValueError("Shared memory fields exceed entity shared memory size")
                    setattr(cls, name, cls._shared_memory_fields_[name])
        imported_keys = {}
        index = 0
        for field in cls._imported_fields_.values():
            keys = list(field.type._flat_keys_(field.data_name))
            if field.default is not None:
                defaults = field.default._to_list_()
                for key, default_value in zip(keys, defaults, strict=True):
                    imported_keys[key] = ImportInfo(index=index, default=default_value)
                    index += 1
            else:
                for key in keys:
                    imported_keys[key] = ImportInfo(index=index, default=None)
                    index += 1
        cls._imported_keys_ = imported_keys
        cls._exported_keys_ = {
            name: i
            for i, name in enumerate(
                key for field in cls._exported_fields_.values() for key in field.type._flat_keys_(field.data_name)
            )
        }
        cls._data_constructor_signature_ = inspect.Signature(
            [inspect.Parameter(name, inspect.Parameter.POSITIONAL_OR_KEYWORD) for name in cls._imported_fields_]
        )
        cls._spawn_signature_ = inspect.Signature(
            [inspect.Parameter(name, inspect.Parameter.POSITIONAL_OR_KEYWORD) for name in cls._memory_fields_]
        )
        cls._post_init_fields()
        cls._field_init_done = True

    @property
    @abstractmethod
    def index(self) -> int:
        """The index of this entity."""
        raise NotImplementedError

    @meta_fn
    def ref(self) -> EntityRef[Self]:
        """Get a reference to this entity.

        Valid both in level data and in callbacks.
        """
        match self._data_:
            case _ArchetypeSelfData():
                return EntityRef[type(self)](index=self.index)  # type: ignore
            case _ArchetypeReferenceData(index=index):
                return EntityRef[type(self)](index=index)  # type: ignore
            case _ArchetypeLevelData():
                result = EntityRef[type(self)](index=-1)  # type: ignore
                result._ref_ = self
                return result
            case _:
                raise RuntimeError("Invalid entity data")

    @classmethod
    def _post_init_fields(cls):
        pass

    @classmethod
    def derive[T](cls: type[T], name: str, is_scored: bool, key: int | float | None = None) -> type[T]:
        """Derive a new archetype class from this archetype.

        Roughly equivalent to returning:
        ```python
        class Derived(cls):
            name = <name>
            is_scored = <is_scored>
            key = <key>  # Only set if key is not None
        ```

        This is used to create a new archetype with the same fields and callbacks, but with a different name and
        whether it is scored. Compared to manually subclassing, this method also enables faster compilation when
        the same base archetype has multiple derived archetypes by compiling callbacks only once for the base archetype.

        Args:
            name: The name of the new archetype.
            is_scored: Whether the new archetype is scored.
            key: A key that can be accessed via the `key` property of the new archetype.

        Returns:
            A new archetype class with the same fields and callbacks as this archetype, but with the given name and
            whether it is scored.
        """
        if getattr(cls, "_is_derived_", False):
            raise RuntimeError("Cannot derive from a derived archetype")
        cls_dict = {
            "name": name,
            "is_scored": is_scored,
            "_is_derived_": True,
            "_derived_base_": cls,
            "__module__": cls.__module__,
        }
        if key is not None:
            if not isinstance(key, (int, float)):
                raise TypeError(f"Key must be an int or float, got {type(key)}")
            cls_dict["key"] = key
        new_cls = type(name, (cls,), cls_dict)
        return new_cls

    @meta_fn
    def _delegate(self, name: str, default: int | float | None = None):
        name = validate_value(name)._as_py_()
        if hasattr(super(), name):
            fn = getattr(super(), name)
            if ctx():
                return compile_and_call(fn)
            else:
                return fn()
        else:
            return default


class PlayArchetype(_BaseArchetype):
    """Base class for play mode archetypes.

    Usage:
        ```python
        class MyArchetype(PlayArchetype):
            # Set to True if the entity is a note and contributes to combo and score
            # Default is False
            is_scored: bool = True

            imported_field: int = imported()
            exported_field: int = exported()
            entity_memory_field: int = entity_memory()
            shared_memory_field: int = shared_memory()

            @callback(order=1)
            def preprocess(self):
                ...
        ```
    """

    _removable_prefix: ClassVar[str] = "Play"

    _supported_callbacks_ = PLAY_CALLBACKS

    is_scored: ClassVar[bool] = False
    """Whether entities of this archetype contribute to combo and score."""

    life: ClassVar[LifeInfo]
    """How entities of this archetype contribute to life depending on judgment.

    Alias for archetype_life, provided for backwards compatibility.
    """

    archetype_life: ClassVar[LifeInfo]
    """How entities of this archetype contribute to life depending on judgment."""

    entity_life: LifeInfo
    """How this specific entity contributes to life depending on judgment.

    This is additive with archetype_life - the total life increment is the sum of both.
    """

    archetype_score_multiplier: ClassVar[float]
    """Score multiplier for entities of this archetype.

    This is additive with other multipliers (except judgment base multipliers).
    """

    entity_score_multiplier: float
    """Score multiplier for this specific entity.

    This is additive with other multipliers (except judgment base multipliers).
    """

    def preprocess(self):
        """Perform upfront processing.

        Runs first when the level is loaded.
        """
        self._delegate("preprocess")

    def spawn_order(self) -> float:
        """Return the spawn order of the entity.

        Runs when the level is loaded after [`preprocess`][sonolus.script.archetype.PlayArchetype.preprocess].
        """
        return self._delegate("spawn_order", default=0.0)

    def should_spawn(self) -> bool:
        """Return whether the entity should be spawned.

        Runs each frame while the entity is the first entity in the spawn queue.
        """
        return self._delegate("should_spawn", default=True)

    def initialize(self):
        """Initialize this entity.

        Runs when this entity is spawned.
        """
        self._delegate("initialize")

    def update_sequential(self):
        """Perform non-parallel actions for this frame.

        Runs first each frame.

        This is where logic affecting shared memory should be placed.
        Other logic should typically be placed in
        [`update_parallel`][sonolus.script.archetype.PlayArchetype.update_parallel]
        for better performance.
        """
        self._delegate("update_sequential")

    def update_parallel(self):
        """Perform parallel actions for this frame.

        Runs after [`touch`][sonolus.script.archetype.PlayArchetype.touch] each frame.

        This is where most gameplay logic should be placed.
        """
        self._delegate("update_parallel")

    def touch(self):
        """Handle user input.

        Runs after [`update_sequential`][sonolus.script.archetype.PlayArchetype.update_sequential] each frame.
        """
        self._delegate("touch")

    def terminate(self):
        """Finalize before despawning.

        Runs when the entity is despawned.
        """
        self._delegate("terminate")

    @property
    @meta_fn
    def despawn(self):
        """Whether the entity should be despawned after this frame.

        Setting this to True will despawn the entity.
        """
        if not ctx():
            raise RuntimeError("Calling despawn is only allowed within a callback")
        match self._data_:
            case _ArchetypeSelfData():
                return _deref(ctx().blocks.EntityDespawn, 0, Num)
            case _:
                raise RuntimeError("Despawn is only accessible from the entity itself")

    @despawn.setter
    @meta_fn
    def despawn(self, value: bool):
        if not ctx():
            raise RuntimeError("Calling despawn is only allowed within a callback")
        match self._data_:
            case _ArchetypeSelfData():
                _deref(ctx().blocks.EntityDespawn, 0, Num)._set_(value)
            case _:
                raise RuntimeError("Despawn is only accessible from the entity itself")

    @property
    @meta_fn
    def _info(self):
        if not ctx():
            raise RuntimeError("Calling info is only allowed within a callback")
        match self._data_:
            case _ArchetypeSelfData():
                return _deref(ctx().blocks.EntityInfo, 0, PlayEntityInfo)
            case _ArchetypeReferenceData(index=index):
                return _deref(ctx().blocks.EntityInfoArray, index * PlayEntityInfo._size_(), PlayEntityInfo)
            case _:
                raise RuntimeError("Info is only accessible from the entity itself")

    @property
    def index(self) -> int:
        """The index of this entity."""
        return self._info.index

    @property
    def is_waiting(self) -> bool:
        """Whether this entity is waiting to be spawned."""
        return self._info.state == 0

    @property
    def is_active(self) -> bool:
        """Whether this entity is active."""
        return self._info.state == 1

    @property
    def is_despawned(self) -> bool:
        """Whether this entity is despawned."""
        return self._info.state == 2

    @property
    @meta_fn
    def result(self) -> PlayEntityInput:
        """The result of this entity.

        Only meaningful for scored entities.
        """
        if not ctx():
            raise RuntimeError("Calling result is only allowed within a callback")
        match self._data_:
            case _ArchetypeSelfData():
                return _deref(ctx().blocks.EntityInput, 0, PlayEntityInput)
            case _:
                raise RuntimeError("Result is only accessible from the entity itself")


class WatchArchetype(_BaseArchetype):
    """Base class for watch mode archetypes.

    Usage:
        ```python
        class MyArchetype(WatchArchetype):
            imported_field: int = imported()
            entity_memory_field: int = entity_memory()
            shared_memory_field: int = shared_memory()

            @callback(order=1)
            def update_sequential(self):
                ...
        ```
    """

    _removable_prefix: ClassVar[str] = "Watch"

    _supported_callbacks_ = WATCH_ARCHETYPE_CALLBACKS

    is_scored: ClassVar[bool] = False
    """Whether entities of this archetype contribute to combo and score."""

    life: ClassVar[LifeInfo]
    """How entities of this archetype contribute to life depending on judgment.

    Alias for archetype_life, provided for backwards compatibility.
    """

    archetype_life: ClassVar[LifeInfo]
    """How entities of this archetype contribute to life depending on judgment."""

    entity_life: LifeInfo
    """How this specific entity contributes to life depending on judgment.

    This is additive with archetype_life - the total life increment is the sum of both.
    """

    archetype_score_multiplier: ClassVar[float]
    """Score multiplier for entities of this archetype.

    This is additive with other multipliers (except judgment base multipliers).
    """

    entity_score_multiplier: float
    """Score multiplier for this specific entity.

    This is additive with other multipliers (except judgment base multipliers).
    """

    def preprocess(self):
        """Perform upfront processing.

        Runs first when the level is loaded.
        """
        self._delegate("preprocess")

    def spawn_time(self) -> float:
        """Return the spawn time of the entity."""
        return self._delegate("spawn_time", default=0.0)

    def despawn_time(self) -> float:
        """Return the despawn time of the entity."""
        return self._delegate("despawn_time", default=0.0)

    def initialize(self):
        """Initialize this entity.

        Runs when this entity is spawned.
        """
        self._delegate("initialize")

    def update_sequential(self):
        """Perform non-parallel actions for this frame.

        Runs first each frame.

        This is where logic affecting shared memory should be placed.
        Other logic should typically be placed in
        [`update_parallel`][sonolus.script.archetype.PlayArchetype.update_parallel] for better performance.
        """
        self._delegate("update_sequential")

    def update_parallel(self):
        """Parallel update callback.

        Runs after [`touch`][sonolus.script.archetype.PlayArchetype.touch] each frame.

        This is where most gameplay logic should be placed.
        """
        self._delegate("update_parallel")

    def terminate(self):
        """Finalize before despawning.

        Runs when the entity is despawned.
        """
        self._delegate("terminate")

    @property
    @meta_fn
    def _info(self):
        if not ctx():
            raise RuntimeError("Calling info is only allowed within a callback")
        match self._data_:
            case _ArchetypeSelfData():
                return _deref(ctx().blocks.EntityInfo, 0, WatchEntityInfo)
            case _ArchetypeReferenceData(index=index):
                return _deref(ctx().blocks.EntityInfoArray, index * WatchEntityInfo._size_(), WatchEntityInfo)
            case _:
                raise RuntimeError("Info is only accessible from the entity itself")

    @property
    def index(self) -> int:
        """The index of this entity."""
        return self._info.index

    @property
    def is_active(self) -> bool:
        """Whether this entity is active."""
        return self._info.state == 1

    @property
    @meta_fn
    def result(self) -> WatchEntityInput:
        """The result of this entity.

        Only meaningful for scored entities.
        """
        if not ctx():
            raise RuntimeError("Calling result is only allowed within a callback")
        match self._data_:
            case _ArchetypeSelfData():
                return _deref(ctx().blocks.EntityInput, 0, WatchEntityInput)
            case _:
                raise RuntimeError("Result is only accessible from the entity itself")

    @classmethod
    def _post_init_fields(cls):
        if cls._exported_fields_:
            raise RuntimeError("Watch archetypes cannot have exported fields")


class PreviewArchetype(_BaseArchetype):
    """Base class for preview mode archetypes.

    Usage:
        ```python
        class MyArchetype(PreviewArchetype):
            imported_field: int = imported()
            entity_memory_field: int = entity_memory()
            shared_memory_field: int = shared_memory()

            @callback(order=1)
            def preprocess(self):
                ...
        ```
    """

    _removable_prefix: ClassVar[str] = "Preview"

    _supported_callbacks_ = PREVIEW_CALLBACKS

    def preprocess(self):
        """Perform upfront processing.

        Runs first when the level is loaded.
        """
        self._delegate("preprocess")

    def render(self):
        """Render the entity.

        Runs after `preprocess`.
        """
        self._delegate("render")

    @property
    @meta_fn
    def _info(self) -> PreviewEntityInfo:
        if not ctx():
            raise RuntimeError("Calling info is only allowed within a callback")
        match self._data_:
            case _ArchetypeSelfData():
                return _deref(ctx().blocks.EntityInfo, 0, PreviewEntityInfo)
            case _ArchetypeReferenceData(index=index):
                return _deref(ctx().blocks.EntityInfoArray, index * PreviewEntityInfo._size_(), PreviewEntityInfo)
            case _:
                raise RuntimeError("Info is only accessible from the entity itself")

    @property
    def index(self) -> int:
        """The index of this entity."""
        return self._info.index

    @classmethod
    def _post_init_fields(cls):
        if cls._exported_fields_:
            raise RuntimeError("Preview archetypes cannot have exported fields")


@meta_fn
def get_archetype_by_name(name: str) -> type[AnyArchetype]:
    """Return the archetype with the given name in the current mode."""
    if not ctx():
        raise RuntimeError("Archetypes by name are only available during compilation.")
    name = validate_value(name)  # type: ignore
    if not name._is_py_():  # type: ignore
        raise TypeError(f"Invalid name: '{name}'")
    name = name._as_py_()  # type: ignore
    if not isinstance(name, str):
        raise TypeError(f"Invalid name: '{name}'")
    archetypes_by_name = ctx().mode_state.archetypes_by_name
    if name not in archetypes_by_name:
        raise KeyError(f"Unknown archetype: '{name}'")
    return archetypes_by_name[name]  # type: ignore


@meta_fn
def entity_info_at(index: int) -> PlayEntityInfo | WatchEntityInfo | PreviewEntityInfo:
    """Retrieve entity info of the entity at the given index.

    Available in play, watch, and preview mode.
    """
    if not ctx():
        raise RuntimeError("Calling entity_info_at is only allowed within a callback")
    match ctx().mode_state.mode:
        case Mode.PLAY:
            return _deref(ctx().blocks.EntityInfoArray, index * PlayEntityInfo._size_(), PlayEntityInfo)
        case Mode.WATCH:
            return _deref(ctx().blocks.EntityInfoArray, index * WatchEntityInfo._size_(), WatchEntityInfo)
        case Mode.PREVIEW:
            return _deref(ctx().blocks.EntityInfoArray, index * PreviewEntityInfo._size_(), PreviewEntityInfo)
        case _:
            raise RuntimeError(f"Entity info is not available in mode '{ctx().mode_state.mode}'")


class PlayEntityInfo(Record):
    index: int
    archetype_id: int
    state: int


class WatchEntityInfo(Record):
    index: int
    archetype_id: int
    state: int


class PreviewEntityInfo(Record):
    index: int
    archetype_id: int


class LifeInfo(Record):
    """How an entity contributes to life."""

    perfect_increment: int
    """Life increment for a perfect judgment."""

    great_increment: int
    """Life increment for a great judgment."""

    good_increment: int
    """Life increment for a good judgment."""

    miss_increment: int
    """Life increment for a miss judgment."""

    def update(
        self,
        perfect_increment: int | None = None,
        great_increment: int | None = None,
        good_increment: int | None = None,
        miss_increment: int | None = None,
    ):
        """Update the life increments."""
        if perfect_increment is not None:
            self.perfect_increment = perfect_increment
        if great_increment is not None:
            self.great_increment = great_increment
        if good_increment is not None:
            self.good_increment = good_increment
        if miss_increment is not None:
            self.miss_increment = miss_increment


ArchetypeLife = LifeInfo
"""Alias for LifeInfo, kept for backwards compatibility."""


class HapticType(IntEnum):
    """The type of haptic feedback for a judgment."""

    NONE = 0
    """No haptic feedback."""

    LIGHT = 1
    """Light haptic feedback."""

    MEDIUM = 2
    """Medium haptic feedback."""

    HEAVY = 3
    """Heavy haptic feedback."""

    LONG = 4
    """Long haptic feedback."""


class PlayEntityInput(Record):
    judgment: Judgment
    accuracy: float
    bucket: Bucket
    bucket_value: float
    haptic: HapticType


class WatchEntityInput(Record):
    target_time: float
    bucket: Bucket
    bucket_value: float


class EntityRef[A: _BaseArchetype](Record):
    """Reference to another entity.

    May be used with `typing.Any` to reference an unknown archetype.

    Usage:
        ```python
        ref = EntityRef[MyArchetype](index=123)

        class MyArchetype(PlayArchetype):
            ref_1: EntityRef[OtherArchetype] = imported()
            ref_2: EntityRef[Any] = imported()
        ```
    """

    index: int

    @classmethod
    def archetype(cls) -> type[A]:
        """Get the archetype type."""
        return cls.type_var_value(A)

    def with_archetype[T: _BaseArchetype](self, archetype: type[T]) -> EntityRef[T]:
        """Return a new reference with the given archetype type."""
        return EntityRef[archetype](index=self.index)

    @meta_fn
    def __eq__(self, other: Any) -> bool:
        if not ctx() and hasattr(self, "_ref_") and hasattr(other, "_ref_"):
            return self._ref_ is other._ref_
        return super().__eq__(other)

    def __hash__(self) -> int:
        if not ctx() and hasattr(self, "_ref_"):
            return hash(id(self._ref_))
        return super().__hash__()

    @meta_fn
    def __bool__(self):
        if ctx():
            static_error("EntityRef cannot be used in a boolean context. Check index directly instead.")
        return True

    @meta_fn
    def get(self, *, check: bool = True) -> A:
        """Get the entity this reference points to.

        Args:
            check: If true, raises an error if the referenced entity is not of this archetype or of a subclass of
                   this archetype. If false, no validation is performed.

        Returns:
            The entity this reference points to.
        """
        assert self.archetype() != Any, (
            "Cannot get entity of unknown (Any) archetype. Use with_archetype() first or use get_as()."
        )
        if ref := getattr(self, "_ref_", None):
            return ref
        return self.archetype().at(self.index, check=check)

    @meta_fn
    def get_as(self, archetype: type[_BaseArchetype]) -> _BaseArchetype:
        """Get the entity as the given archetype type."""
        if getattr(archetype, "_ref_", None):
            raise TypeError("Using get_as in level data is not supported.")
        return self.with_archetype(archetype).get()

    def archetype_matches(self, strict: bool = False) -> bool:
        """Check if entity at the index is of this archetype.

        Args:
            strict: If true, only returns true if the entity is exactly of this archetype. If false, also returns true
                    if the entity is of a subclass of this archetype.

        Returns:
            Whether the entity at the given index is of this archetype.
        """
        assert self.archetype() != Any, (
            "Cannot use archetype_matches with unknown (Any) archetype. Use with_archetype() first."
        )
        return self.archetype().is_at(self.index, strict=strict)

    def _to_list_(self, level_refs: dict[Any, str] | None = None) -> list[DataValue | str]:
        ref = getattr(self, "_ref_", None)
        if ref is None:
            return Num._accept_(self.index)._to_list_()
        else:
            if level_refs is None:
                raise RuntimeError("Unexpected missing level_refs")
            if ref not in level_refs:
                raise KeyError("Reference to entity not in level data")
            return [level_refs[ref]]

    def _copy_from_(self, value: Any):
        super()._copy_from_(value)
        if hasattr(value, "_ref_"):
            self._ref_ = value._ref_

    @classmethod
    def _accepts_(cls, value: Any) -> bool:
        return (
            super()._accepts_(value)
            or (cls._type_args_ and cls.archetype() is Any and isinstance(value, EntityRef))
            or (issubclass(type(value), EntityRef) and issubclass(value.archetype(), cls.archetype()))
        )

    @classmethod
    def _accept_(cls, value: Any) -> Self:
        if not cls._accepts_(value):
            raise TypeError(f"Expected {cls}, got {type(value)}")
        result = value.with_archetype(cls.archetype())
        if hasattr(value, "_ref_"):
            result._ref_ = value._ref_
        return result

    @classmethod
    def _validate_parameterized_(cls):
        if not issubclass(cls.archetype(), _BaseArchetype) and cls.archetype() is not Any:
            raise TypeError("EntityRef type parameter must be an Archetype or Any")


class StandardArchetypeName(StrEnum):
    """Standard archetype names."""

    BPM_CHANGE = "#BPM_CHANGE"
    """Bpm change marker"""

    TIMESCALE_CHANGE = "#TIMESCALE_CHANGE"
    """Timescale change marker"""

    TIMESCALE_GROUP = "#TIMESCALE_GROUP"
    """Entity referenced by the timescale changes in a group"""


type AnyArchetype = PlayArchetype | WatchArchetype | PreviewArchetype
"""Union of all archetype types."""


class StandardImportName:
    """Standard import names for Archetype fields.

    Usage:
        ```python
        class MyArchetype(WatchArchetype):
            judgment: int = imported(name=StandardImportName.JUDGMENT)
        ```
    """

    BEAT = "#BEAT"
    """The beat of the entity."""

    BPM = "#BPM"
    """The bpm, for bpm change markers."""

    TIMESCALE = "#TIMESCALE"
    """The timescale, for timescale change markers."""

    TIMESCALE_SKIP = "#TIMESCALE_SKIP"
    """The scaled time to skip, for timescale change markers."""

    TIMESCALE_GROUP = "#TIMESCALE_GROUP"
    """The timescale group, for timescale change markers."""

    TIMESCALE_EASE = "#TIMESCALE_EASE"
    """The timescale ease type, for timescale change markers."""

    JUDGMENT = "#JUDGMENT"
    """The judgment of the entity.

    Automatically set in watch mode for archetypes with a corresponding scored play mode archetype.
    """

    ACCURACY = "#ACCURACY"
    """The accuracy of the entity.

    Automatically set in watch mode for archetypes with a corresponding scored play mode archetype.
    """


class StandardImport:
    """Standard import annotations for Archetype fields.

    Usage:
        ```python
        class MyArchetype(WatchArchetype):
            judgment: StandardImport.JUDGMENT
        ```
    """

    BEAT = Annotated[float, imported(name=StandardImportName.BEAT)]
    """The beat of the entity."""

    BPM = Annotated[float, imported(name=StandardImportName.BPM)]
    """The bpm, for bpm change markers."""

    TIMESCALE = Annotated[float, imported(name=StandardImportName.TIMESCALE)]
    """The timescale, for timescale change markers."""

    TIMESCALE_SKIP = Annotated[float, imported(name=StandardImportName.TIMESCALE_SKIP)]
    """The scaled time to skip, for timescale change markers."""

    TIMESCALE_GROUP = Annotated[EntityRef[Any], imported(name=StandardImportName.TIMESCALE_GROUP)]
    """The timescale group, for timescale change markers."""

    TIMESCALE_EASE = Annotated[TimescaleEase, imported(name=StandardImportName.TIMESCALE_EASE)]
    """The timescale ease type, for timescale change markers."""

    JUDGMENT = Annotated[Judgment, imported(name=StandardImportName.JUDGMENT)]
    """The judgment of the entity.

    Automatically set in watch mode for archetypes with a corresponding scored play mode archetype.
    """
    ACCURACY = Annotated[float, imported(name=StandardImportName.ACCURACY)]
    """The accuracy of the entity.

    Automatically set in watch mode for archetypes with a corresponding scored play mode archetype.
    """



from __future__ import annotations

from abc import ABCMeta
from collections.abc import Iterable
from typing import Any, Literal, Self, final

from sonolus.backend.ir import IRConst, IRSet
from sonolus.backend.place import BlockPlace
from sonolus.script.array_like import ArrayLike, get_positive_index
from sonolus.script.debug import assert_unreachable
from sonolus.script.internal.context import ctx
from sonolus.script.internal.error import InternalError
from sonolus.script.internal.generic import GenericValue
from sonolus.script.internal.impl import validate_value
from sonolus.script.internal.meta_fn import meta_fn, perf_meta_fn
from sonolus.script.internal.value import BackingSource, DataValue, Value
from sonolus.script.num import Num

Dim = Literal
"""Shorthand for `typing.Literal` intended for use in array dimensions for type checker compatibility."""


class ArrayMeta(ABCMeta):
    @meta_fn
    def __pos__[T](cls: type[T]) -> T:
        """Create a zero-initialized array instance."""
        return cls._zero_()


@final
class Array[T, Size](GenericValue, ArrayLike[T], metaclass=ArrayMeta):
    """A fixed size array of values.

    Supports negative indexes.

    Usage:
        ```python
        array_1 = Array(1, 2, 3)
        array_2 = Array[int, 0]()
        array_3 = +Array[int, 3]  # Create a zero-initialized array
        array_4 = +Array[int, Dim[3]]  # Alternative syntax for compliance with type checkers
        array_5 = +array_1  # Create a copy of array_1
        ```
    """

    _value: list[T] | BlockPlace | BackingSource

    @classmethod
    def element_type(cls) -> type[T] | type[Value]:
        """Return the type of elements in this array type."""
        return cls.type_var_value(T)

    @classmethod
    def size(cls) -> int:
        """Return the size of this array type.

        On instances, use `len(array)` instead.
        """
        return cls.type_var_value(Size)

    def __new__(cls, *args: T) -> Array[T, Any]:
        if cls._type_args_ is None:
            values = [validate_value(arg) for arg in args]
            types = {type(value) for value in values}
            if len(types) == 0:
                raise ValueError(
                    f"{cls.__name__} constructor should be used with at least one value if type is not specified"
                )
            if len(types) > 1:
                raise TypeError(f"{cls.__name__} constructor should be used with values of the same type, got {types}")
            parameterized_cls = cls[types.pop(), len(args)]
        else:
            values = [cls.element_type()._accept_(arg) for arg in args]
            if len(args) != cls.size():
                raise ValueError(f"{cls.__name__} constructor should be used with {cls.size()} values, got {len(args)}")
            parameterized_cls = cls
        if ctx():
            place = ctx().alloc(size=parameterized_cls._size_())
            result: parameterized_cls = parameterized_cls._from_place_(place)
            result._copy_from_(parameterized_cls._with_value(values))
            return result
        else:
            return parameterized_cls._with_value([value._copy_() for value in values])

    def __init__(self, *args: T):
        super().__init__()

    @classmethod
    def _with_value(cls, value) -> Array[T, Size]:
        result = object.__new__(cls)
        result._value = value
        return result

    @classmethod
    def _size_(cls) -> int:
        return cls.size() * cls.element_type()._size_()

    @classmethod
    def _is_value_type_(cls) -> bool:
        return False

    @classmethod
    def _from_backing_source_(cls, source: BackingSource) -> Array[T, Size]:
        return cls._with_value(source)

    @classmethod
    def _from_place_(cls, place: BlockPlace) -> Array[T, Size]:
        return cls._with_value(place)

    @classmethod
    def _accepts_(cls, value: Any) -> bool:
        return isinstance(value, cls)

    @classmethod
    def _accept_(cls, value: Any) -> Array[T, Size]:
        if not cls._accepts_(value):
            raise TypeError(f"Cannot accept value {value} as {cls.__name__}")
        return value

    def _is_py_(self) -> bool:
        return isinstance(self._value, list)

    def _as_py_(self) -> Any:
        if not self._is_py_():
            raise ValueError("Not a python value")
        return self

    @classmethod
    def _from_list_(cls, values: Iterable[DataValue]) -> Array[T, Size]:
        iterator = iter(values)
        return cls(*(cls.element_type()._from_list_(iterator) for _ in range(cls.size())))

    def _to_list_(self, level_refs: dict[Any, str] | None = None) -> list[DataValue | str]:
        match self._value:
            case list():
                return [entry for value in self._value for entry in value._to_list_(level_refs)]
            case BlockPlace():
                return [
                    entry
                    for i in range(self.size())
                    for entry in self.element_type()
                    ._from_place_(self._value.add_offset(i * self.element_type()._size_()))
                    ._to_list_()
                ]
            case backing_source if callable(backing_source):
                return [backing_source(IRConst(i)) for i in range(self.size() * self.element_type()._size_())]
            case _:
                assert_unreachable()

    @classmethod
    def _flat_keys_(cls, prefix: str) -> list[str]:
        return [entry for i in range(cls.size()) for entry in cls.element_type()._flat_keys_(f"{prefix}[{i}]")]

    def _get_(self) -> Array[T, Size]:
        return self

    def _set_(self, value: Any):
        raise TypeError("Array does not support _set_")

    def _copy_from_(self, value: Any):
        if not isinstance(value, type(self)):
            raise TypeError("Cannot copy from different type")
        for i in range(self.size()):
            self[i] = value[i]

    def _copy_(self) -> Array[T, Size]:
        if ctx():
            place = ctx().alloc(size=self._size_())
            result: Self = self._from_place_(place)
            result._copy_from_(self)
            return result
        else:
            assert isinstance(self._value, list)
            return self._with_value([value._copy_() for value in self._value])

    @classmethod
    def _alloc_(cls) -> Array[T, Size]:
        if ctx():
            place = ctx().alloc(size=cls._size_())
            return cls._from_place_(place)
        else:
            return cls._with_value([cls.element_type()._alloc_() for _ in range(cls.size())])

    @classmethod
    def _zero_(cls) -> Array[T, Size]:
        if ctx():
            place = ctx().alloc(size=cls._size_())
            result: Self = cls._from_place_(place)
            ctx().add_statements(*[IRSet(place.add_offset(i), IRConst(0)) for i in range(cls._size_())])
            return result
        else:
            return cls._with_value([cls.element_type()._zero_() for _ in range(cls.size())])

    @perf_meta_fn
    def __len__(self):
        return self.size()

    @meta_fn
    def __getitem__(self, index: int) -> T:
        return self.get_unchecked(get_positive_index(index, self.size()))

    @meta_fn
    def get_unchecked(self, index: Num) -> T:
        """Get the element at the given index possibly without bounds checking.

        The compiler may still determine that the index is out of bounds and throw an error, but it may skip these
        checks at runtime.

        Args:
            index: The index to get.

        Returns:
            The element at the given index.
        """
        index = Num._accept_(index)
        if index._is_py_() and 0 <= index._as_py_() < self.size():
            const_index = index._as_py_()
            if isinstance(const_index, float) and not const_index.is_integer():
                raise ValueError("Array index must be an integer")
            const_index = int(const_index)
            if isinstance(self._value, list):
                if ctx():
                    return self._value[const_index]._get_()
                else:
                    return self._value[const_index]._get_()._as_py_()
            elif isinstance(self._value, BlockPlace):
                return (
                    self.element_type()
                    ._from_place_(self._value.add_offset(const_index * self.element_type()._size_()))
                    ._get_()
                )
            elif callable(self._value):
                return self.element_type()._from_backing_source_(
                    lambda offset: self._value((Num(offset) + Num(const_index * self.element_type()._size_())).ir())
                    # type: ignore
                )
            else:
                raise InternalError("Unexpected array value")
        else:
            if not ctx():
                raise InternalError("Unexpected non-constant index")
            if isinstance(self._value, list | BlockPlace):
                base = ctx().rom[tuple(self._to_list_())] if isinstance(self._value, list) else self._value
                place = BlockPlace(
                    block=base.block,
                    index=(Num(base.index) + index * self.element_type()._size_()).index(),
                    offset=base.offset,
                )
                return self.element_type()._from_place_(place)._get_()
            elif callable(self._value):
                base_offset = index * Num(self.element_type()._size_())
                return self.element_type()._from_backing_source_(
                    lambda offset: self._value((Num(offset) + base_offset).ir())  # type: ignore
                )
            else:
                raise InternalError("Unexpected array value")

    @meta_fn
    def __setitem__(self, index: int, value: T):
        self.set_unchecked(get_positive_index(index, self.size()), value)

    @meta_fn
    def set_unchecked(self, index: Num, value: T):
        """Set the element at the given index possibly without bounds checking.

        The compiler may still determine that the index is out of bounds and throw an error, but it may skip these
        checks at runtime.

        Args:
            index: The index to set.
            value: The value to set.
        """
        index = Num._accept_(index)
        value = self.element_type()._accept_(value)
        if ctx():
            if isinstance(self._value, list):
                raise ValueError("Cannot mutate a compile time constant array")
            elif isinstance(self._value, BlockPlace):
                base = self._value
                place = (
                    base.add_offset(int(index._as_py_()) * self.element_type()._size_())
                    if index._is_py_()
                    else BlockPlace(
                        block=base.block,
                        index=(Num(base.index) + index * self.element_type()._size_()).index(),
                        offset=base.offset,
                    )
                )
                dst = self.element_type()._from_place_(place)
            elif callable(self._value):
                base_offset = index * Num(self.element_type()._size_())
                dst = self.element_type()._from_backing_source_(
                    lambda offset: self._value((Num(offset) + base_offset).ir())  # type: ignore
                )
            else:
                raise InternalError("Unexpected array value")
            if self.element_type()._is_value_type_():
                dst._set_(value)
            else:
                dst._copy_from_(value)
        else:
            if not isinstance(self._value, list):
                raise InternalError("Unexpected mutation of non compile time constant array")
            const_index = index._as_py_()
            if isinstance(const_index, float) and not const_index.is_integer():
                raise ValueError("Array index must be an integer")
            const_index = int(const_index)
            if not 0 <= const_index < self.size():
                raise IndexError("Array index out of range")
            dst = self._value[const_index]
            if self.element_type()._is_value_type_():
                dst._set_(value)
            else:
                dst._copy_from_(value)

    def __eq__(self, other):
        if not isinstance(other, ArrayLike):
            return False
        if len(self) != len(other):
            return False
        i = 0
        while i < self.size():
            if self.get_unchecked(i) != other.get_unchecked(i):
                return False
            i += 1
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(tuple(self[i] for i in range(self.size())))

    def __str__(self):
        if isinstance(self._value, BlockPlace) or callable(self._value):
            return f"{type(self).__name__}({self._value}...)"
        else:
            return f"{type(self).__name__}({', '.join(str(self[i]) for i in range(self.size()))})"

    def __repr__(self):
        if isinstance(self._value, BlockPlace) or callable(self._value):
            return f"{type(self).__name__}({self._value}...)"
        else:
            return f"{type(self).__name__}({', '.join(repr(self[i]) for i in range(self.size()))})"

    @meta_fn
    def __pos__(self) -> Array[T, Size]:
        """Return a copy of the array."""
        return self._copy_()
